#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from datetime import timedelta
from dateutil import rrule
from icalendar import Calendar
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from karbor import exception
from karbor.i18n import _
from karbor.services.operationengine.engine.triggers.timetrigger import \
    timeformats


RATE = 2
FREQ_TO_KWARGS = [{'days': RATE*366},
                  {'days': RATE*31},
                  {'days': RATE*7},
                  {'days': RATE},
                  {'hours': RATE},
                  {'minutes': RATE},
                  {'seconds': RATE}]

RREQ_MAP = {"YEARLY": 0,
            "MONTHLY": 1,
            "WEEKLY": 2,
            "DAILY": 3,
            "HOURLY": 4,
            "MINUTELY": 5,
            "SECONDLY": 6}


class ICal(timeformats.TimeFormat):
    """icalendar."""

    def __init__(self, start_time, pattern):
        super(ICal, self).__init__(start_time, pattern)
        cal = Calendar.from_ical(self._decode_calendar_pattern(pattern))
        vevent = cal.walk('VEVENT')[0]
        self.dtstart = start_time
        self.min_freq = self._get_min_freq(vevent)
        self.rrule_obj = self._get_rrule_obj(vevent, start_time)

    @staticmethod
    def _decode_calendar_pattern(pattern):
        try:
            pattern.index('\\')
            pattern_dict = jsonutils.loads('{"pattern": "%s"}' % pattern)
            return pattern_dict["pattern"]
        except Exception:
            return pattern

    @staticmethod
    def _get_rrule_obj(vevent, dtstart):
        rrules = vevent.get('RRULE')
        rrule_list = rrules if isinstance(rrules, list) else [rrules]
        rrule_str = os.linesep.join(recur.to_ical().decode("utf-8")
                                    for recur in rrule_list)
        return rrule.rrulestr(rrule_str, dtstart=dtstart, cache=False)

    @staticmethod
    def _get_min_freq(vevent):
        recur = vevent.decoded("RRULE")
        recur_list = recur if isinstance(recur, list) else [recur]
        freq_list = []
        for recur in recur_list:
            for freq in recur.get("FREQ"):
                freq_list.append(RREQ_MAP[freq.upper()])
        return max(freq_list)

    @classmethod
    def check_time_format(cls, pattern):
        """Check time format

        :param pattern: The pattern of the icalendar time
        """
        try:
            cal_obj = Calendar.from_ical(cls._decode_calendar_pattern(pattern))
        except Exception:
            msg = (_("The trigger pattern(%s) is invalid") % pattern)
            raise exception.InvalidInput(msg)

        try:
            vevent = cal_obj.walk('VEVENT')[0]
        except Exception:
            msg = (_("The trigger pattern(%s) must include less than one "
                     "VEVENT component") % pattern)
            raise exception.InvalidInput(msg)

        try:
            vevent.decoded('RRULE')
        except Exception:
            msg = (_("The first VEVENT component of trigger pattern(%s) must "
                     "include less than one RRULE property") % pattern)
            raise exception.InvalidInput(msg)

    def compute_next_time(self, current_time):
        """Compute next time

        :param current_time: the time before the next time
        :return datetime or None
        """
        next_time = self.rrule_obj.after(current_time)
        return next_time if next_time else None

    def get_min_interval(self):
        """Get minimum interval of two adjacent time points

        :return int(seconds) or None
        """
        gen = self.rrule_obj
        kwargs = FREQ_TO_KWARGS[self.min_freq]
        endtime = self.dtstart + timedelta(**kwargs)

        deltas = []
        t0 = None
        for dt in gen:
            if dt > endtime:
                break
            t1 = t0
            t0 = dt
            if t1 is None or t0 is None or dt <= self.dtstart:
                continue
            delta = timeutils.delta_seconds(t1, t0)
            if delta:
                deltas.append(delta)
        if len(deltas):
            return min(deltas)
        else:
            return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from karbor import objects

db_plan = {
    'id': '1',
    'name': 'My 3 tier application',
    'provider_id': 'efc6a88b-9096-4bb6-8634-cda182a6e12a',
    'status': 'started',
    'project_id': '39bb894794b741e982bd26144d2949f6',
    'resources': [],
    'parameters': '{"OS::Nova::Server": {"consistency": "os"}}',
}


def fake_db_plan(**updates):
    for name, field in objects.Plan.fields.items():
        if name in db_plan:
            continue
        if field.nullable:
            db_plan[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_plan[name] = field.default
        else:
            raise Exception('fake_db_plan needs help with %s.' % name)

    if updates:
        db_plan.update(updates)

    return db_plan



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.common import constants
from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects


class CheckpointsTest(karbor_base.KarborBaseTest):
    """Test Checkpoints operation """
    def setUp(self):
        super(CheckpointsTest, self).setUp()
        self.provider_id = self.provider_id_os

    def test_checkpoint_create(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        volume_parameter_key = "OS::Cinder::Volume#{id}".format(id=volume.id)
        backup_name = "volume-backup-{id}".format(id=volume.id)
        parameters = {
            "OS::Cinder::Volume": {
                "backup_mode": "full",
                "force": False
            },
            volume_parameter_key: {
                "backup_name": backup_name
            }
        }
        plan.create(self.provider_id_os, [volume, ],
                    parameters=parameters)

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        search_opts = {"volume_id": volume.id}
        backups = self.cinder_client.backups.list(search_opts=search_opts)
        self.assertEqual(1, len(backups))

        search_opts = {"name": backup_name}
        backups = self.cinder_client.backups.list(search_opts=search_opts)
        self.assertEqual(1, len(backups))

    def test_checkpoint_delete(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [volume, ])

        checkpoint = objects.Checkpoint()
        checkpoint.create(self.provider_id, plan.id, timeout=2400)
        checkpoint_item = self.karbor_client.checkpoints.get(self.provider_id,
                                                             checkpoint.id)
        self.assertEqual(constants.CHECKPOINT_STATUS_AVAILABLE,
                         checkpoint_item.status)

        checkpoint.close()
        items = self.karbor_client.checkpoints.list(self.provider_id)
        ids = [item.id for item in items]
        self.assertTrue(checkpoint.id not in ids)

    def test_checkpoint_list(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [volume, ])

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        items = self.karbor_client.checkpoints.list(self.provider_id)
        ids = [item.id for item in items]
        self.assertTrue(checkpoint.id in ids)

    def test_checkpoint_get(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [volume, ])

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        # sanity
        checkpoint_item = self.karbor_client.checkpoints.get(self.provider_id,
                                                             checkpoint.id)
        self.assertEqual(constants.CHECKPOINT_STATUS_AVAILABLE,
                         checkpoint_item.status)
        self.assertEqual(checkpoint.id, checkpoint_item.id)

    def test_server_attached_volume_only_protect_server(self):
        """Test checkpoint for server with attached volume

        Test checkpoint for server which has attached one volume,
        but only add server in protect source
        """
        volume = self.store(objects.Volume())
        volume.create(1)
        server = self.store(objects.Server())
        server.create()
        server.attach_volume(volume.id)

        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [server, ])

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        items = self.karbor_client.checkpoints.list(self.provider_id)
        ids = [item.id for item in items]
        self.assertTrue(checkpoint.id in ids)
        search_opts = {"volume_id": volume.id}
        backups = self.cinder_client.backups.list(search_opts=search_opts)
        self.assertEqual(1, len(backups))
        server.detach_volume(volume.id)

    def test_server_attached_volume_protect_both(self):
        """Test checkpoint for server with attached volume

        Test checkpoint for server which has attached one volume,
        and add server and volume both in protect source
        """
        volume = self.store(objects.Volume())
        volume.create(1)
        server = self.store(objects.Server())
        server.create()
        server.attach_volume(volume.id)

        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [server, volume])

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        items = self.karbor_client.checkpoints.list(self.provider_id)
        ids = [item.id for item in items]
        self.assertTrue(checkpoint.id in ids)
        search_opts = {"volume_id": volume.id}
        backups = self.cinder_client.backups.list(search_opts=search_opts)
        self.assertEqual(1, len(backups))
        server.detach_volume(volume.id)

    def test_server_boot_from_volume_with_attached_volume(self):
        """Test checkpoint for server with a bootable volume

        Test checkpoint for server which has booted form one bootable
        volume.
        """
        bootable_volume = self.store(objects.Volume())
        bootable_volume_id = bootable_volume.create(1, create_from_image=True)
        volume = self.store(objects.Volume())
        volume.create(1)
        server = self.store(objects.Server())
        server.create(volume=bootable_volume_id)
        server.attach_volume(volume.id)

        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [server, ])

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        items = self.karbor_client.checkpoints.list(self.provider_id)
        ids = [item.id for item in items]
        self.assertTrue(checkpoint.id in ids)
        search_opts = {"volume_id": volume.id}
        backups = self.cinder_client.backups.list(search_opts=search_opts)
        self.assertEqual(1, len(backups))
        search_opts = {"volume_id": bootable_volume_id}
        bootable_backups = self.cinder_client.backups.list(
            search_opts=search_opts)
        self.assertEqual(1, len(bootable_backups))
        server.detach_volume(volume.id)

    def test_checkpoint_share_projection(self):
        share = self.store(objects.Share())
        share.create("NFS", 1)
        plan = self.store(objects.Plan())

        share_parameter_key = "OS::Manila::Share#{id}".format(
            id=share.id)
        snapshot_name = "share-snapshot-{id}".format(id=share.id)
        parameters = {
            "OS::Manila::Share": {
                "force": False
            },
            share_parameter_key: {
                "snapshot_name": snapshot_name
            }
        }
        plan.create(self.provider_id_os, [share, ],
                    parameters=parameters)

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id, plan.id, timeout=2400)

        search_opts = {"share_id": share.id}
        snapshots = self.manila_client.share_snapshots.list(
            search_opts=search_opts)
        self.assertEqual(1, len(snapshots))

        search_opts = {"name": snapshot_name}
        backups = self.manila_client.share_snapshots.list(
            search_opts=search_opts)
        self.assertEqual(1, len(backups))

    def test_checkpoint_volume_snapshot(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        volume_parameter_key = "OS::Cinder::Volume#{id}".format(id=volume.id)
        snapshot_name = "volume-snapshot-{id}".format(id=volume.id)
        parameters = {
            "OS::Cinder::Volume": {
                "force": False
            },
            volume_parameter_key: {
                "snapshot_name": snapshot_name
            }
        }
        plan.create(self.provider_id_os_volume_snapshot, [volume, ],
                    parameters=parameters)

        checkpoint = self.store(objects.Checkpoint())
        checkpoint.create(self.provider_id_os_volume_snapshot, plan.id,
                          timeout=2400)

        search_opts = {"volume_id": volume.id}
        snapshots = self.cinder_client.volume_snapshots.list(
            search_opts=search_opts)
        self.assertEqual(1, len(snapshots))

        search_opts = {"name": snapshot_name}
        snapshots = self.cinder_client.volume_snapshots.list(
            search_opts=search_opts)
        self.assertEqual(1, len(snapshots))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os
import re
import six

from karbor import exception
from karbor.i18n import _


@six.add_metaclass(abc.ABCMeta)
class LeasePlugin(object):
    @abc.abstractmethod
    def acquire_lease(self):
        pass

    @abc.abstractmethod
    def renew_lease(self):
        pass

    @abc.abstractmethod
    def check_lease_validity(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class BankPlugin(object):
    def __init__(self, config=None):
        super(BankPlugin, self).__init__()
        self._config = config

    @abc.abstractmethod
    def update_object(self, key, value):
        return

    @abc.abstractmethod
    def get_object(self, key):
        return

    @abc.abstractmethod
    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        return

    @abc.abstractmethod
    def delete_object(self, key):
        return

    @abc.abstractmethod
    def get_owner_id(self):
        return


def validate_key(key):
    pass


def validate_dir(key):
    pass


class Bank(object):
    _KEY_VALIDATION = re.compile('^[A-Za-z0-9/_.\-@]+(?<!/)$')
    _KEY_DOT_VALIDATION = re.compile('/\.{1,2}(/|$)')

    def __init__(self, plugin):
        super(Bank, self).__init__()
        self._plugin = plugin

    def _normalize_key(self, key):
        """Normalizes the key

        To prevent small errors regarding path joining we define that
        banks use path normalization similar to file systems.

        This means that all paths are relative to '/' and that
        '/path//dir' == '/path/dir'
        """
        res = os.path.normpath(key)
        if not res.startswith("/"):
            res = "/" + res

        if key.endswith("/"):
            res += "/"

        return res

    @classmethod
    def _validate_key(cls, key):
        if not isinstance(key, six.string_types):
            raise exception.InvalidParameterValue(
                err=_('Key must be a string')
            )

        if not cls._KEY_VALIDATION.match(key):
            raise exception.InvalidParameterValue(
                err=_('Only alphanumeric, underscore, dash, dots, at signs, '
                      'and slashes are allowed. Key: "%s"') % key
            )

        if cls._KEY_DOT_VALIDATION.match(key):
            raise exception.InvalidParameterValue(
                err=_('Invalid parameter: must not contain "." or ".." parts')
            )

    def update_object(self, key, value):
        self._validate_key(key)
        return self._plugin.update_object(self._normalize_key(key), value)

    def get_object(self, key):
        self._validate_key(key)
        return self._plugin.get_object(self._normalize_key(key))

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        if not prefix:
            prefix = "/"

        norm_prefix = self._normalize_key(prefix)

        return self._plugin.list_objects(
            prefix=norm_prefix,
            limit=limit,
            marker=marker,
            sort_dir=sort_dir
        )

    def delete_object(self, key):
        self._validate_key(key)
        return self._plugin.delete_object(self._normalize_key(key))

    def get_sub_section(self, section, is_writable=True):
        return BankSection(self, section, is_writable)

    @property
    def is_writeable(self):
        return True

    def get_owner_id(self):
        return self._plugin.get_owner_id()


class BankSection(object):
    """Bank Section compartmentalizes a section of a bank.

    Bank section is used when an object wants to pass a section of
    a bank to another entity and make sure it is only capable of
    accessing part of it.
    """
    _SECTION_VALIDATION = re.compile('^/?[A-Za-z0-9/_.\-@]*/?$')
    _SECTION_DOT_VALIDATION = re.compile('/\.{1,2}(/|$)')

    def __init__(self, bank, section, is_writable=True):
        super(BankSection, self).__init__()
        self._validate_section(section)

        self._bank = bank
        self._prefix = os.path.normpath(section)
        if not self._prefix.startswith('/'):
            self._prefix = '/' + self._prefix
        if not self._prefix.endswith('/'):
            self._prefix += '/'
        self._is_writable = is_writable

    def get_sub_section(self, prefix, is_writable=True):
        if is_writable and not self._is_writable:
            raise exception.BankReadonlyViolation()

        return BankSection(self._bank, self._prefix + '/' + prefix,
                           self._is_writable)

    @classmethod
    def _validate_section(cls, section):
        if not section:
            raise exception.InvalidParameterValue(
                err=_('Empty section')
            )

        if not cls._SECTION_VALIDATION.match(section):
            raise exception.InvalidParameterValue(
                err=_('Invalid section. Must begin and end with a slash, '
                      'and contain valid characters')
            )

        if cls._SECTION_DOT_VALIDATION.match(section):
            raise exception.InvalidParameterValue(
                err=_('Invalid parameter: must not contain "." or ".." parts')
            )

    @property
    def is_writable(self):
        return self._is_writable

    def _prepend_prefix(self, key):
        if not isinstance(key, six.string_types):
            raise exception.InvalidParameterValue(
                err=_('Key must be a string')
            )
        full_key = self._prefix + key
        res = os.path.normpath(full_key)
        if full_key.endswith('/'):
            res += '/'
        return res

    def _validate_writable(self):
        if not self.is_writable:
            raise exception.BankReadonlyViolation()

    def update_object(self, key, value):
        self._validate_writable()
        return self._bank.update_object(
            self._prepend_prefix(key),
            value,
        )

    def get_object(self, key):
        return self._bank.get_object(
            self._prepend_prefix(key),
        )

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        if not prefix:
            prefix = self._prefix
        else:
            prefix = self._prepend_prefix(prefix)

        if marker is not None:
            marker = self._prepend_prefix(marker)

        return [
            key[len(self._prefix):]
            for key in self._bank.list_objects(
                prefix,
                limit,
                marker,
                sort_dir
            )
        ]

    def delete_object(self, key):
        self._validate_writable()
        return self._bank.delete_object(
            self._prepend_prefix(key),
        )

    def get_owner_id(self):
        return self._bank.get_owner_id()

    @property
    def bank(self):
        return self._bank



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from karbor.common import constants
from karbor import context
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.services.operationengine.operations import base


class ProtectOperation(base.Operation):
    """Protect operation."""

    OPERATION_TYPE = "protect"

    def check_operation_definition(self, operation_definition):
        provider_id = operation_definition.get("provider_id")
        if not provider_id or not uuidutils.is_uuid_like(provider_id):
            reason = _("Provider_id is invalid")
            raise exception.InvalidOperationDefinition(reason=reason)

        plan_id = operation_definition.get("plan_id")
        if not plan_id or not uuidutils.is_uuid_like(plan_id):
            reason = _("Plan_id is invalid")
            raise exception.InvalidOperationDefinition(reason=reason)

        plan = objects.Plan.get_by_id(context.get_admin_context(), plan_id)
        if provider_id != plan.provider_id:
            reason = _("Provider_id is invalid")
            raise exception.InvalidOperationDefinition(reason=reason)

    def _execute(self, operation_definition, param):
        log_ref = self._create_operation_log(param)
        self._run(operation_definition, param, log_ref)

    def _resume(self, operation_definition, param, log_ref):
        self._run(operation_definition, param, log_ref)

    def _run(self, operation_definition, param, log_ref):
        client = self._create_karbor_client(
            param.get("user_id"), param.get("project_id"))
        try:
            client.checkpoints.create(operation_definition.get("provider_id"),
                                      operation_definition.get("plan_id"))
        except Exception:
            state = constants.OPERATION_EXE_STATE_FAILED
        else:
            state = constants.OPERATION_EXE_STATE_SUCCESS

        self._update_log_when_operation_finished(log_ref, state)



#!/usr/bin/env python
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
  CLI interface for karbor management.
"""

from __future__ import print_function


import os
import sys

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import migration
from oslo_log import log as logging

from karbor import i18n
i18n.enable_lazy()

# Need to register global_opts
from karbor.common import config  # noqa
from karbor import context
from karbor import db
from karbor.db import migration as db_migration
from karbor.db.sqlalchemy import api as db_api
from karbor.i18n import _
from karbor import objects
from karbor import utils
from karbor import version


CONF = cfg.CONF


# Decorators for actions
def args(*args, **kwargs):
    def _decorator(func):
        func.__dict__.setdefault('args', []).insert(0, (args, kwargs))
        return func
    return _decorator


class DbCommands(object):
    """Class for managing the database."""

    @args('version', nargs='?', default=None, type=int,
          help='Database version')
    def sync(self, version=None):
        """Sync the database up to the most recent version."""
        if version is not None and version > db.MAX_INT:
            print(_('Version should be less than or equal to '
                    '%(max_version)d.') % {'max_version': db.MAX_INT})
            sys.exit(1)
        try:
            return db_migration.db_sync(version)
        except db_exc.DbMigrationError as ex:
            print("Error during database migration: %s" % ex)
            sys.exit(1)

    def version(self):
        """Print the current database version."""
        print(db_migration.MIGRATE_REPO_PATH)
        print(migration.db_version(db_api.get_engine(),
                                   db_migration.MIGRATE_REPO_PATH,
                                   db_migration.INIT_VERSION))

    @args('age_in_days', type=int,
          help='Purge deleted rows older than age in days')
    def purge(self, age_in_days):
        """Purge deleted rows older than a given age from karbor tables."""
        age_in_days = int(age_in_days)
        if age_in_days <= 0:
            print(_("Must supply a positive, non-zero value for age"))
            sys.exit(1)
        ctxt = context.get_admin_context()

        try:
            db.purge_deleted_rows(ctxt, age_in_days)
        except Exception as e:
            print(_("Purge command failed, check karbor-manage "
                    "logs for more details. %s") % e)
            sys.exit(1)


class VersionCommands(object):
    """Class for exposing the codebase version."""

    def list(self):
        print(version.version_string())

    def __call__(self):
        self.list()


class ConfigCommands(object):
    """Class for exposing the flags defined by flag_file(s)."""

    @args('param', nargs='?', default=None,
          help='Configuration parameter to display (default: %(default)s)')
    def list(self, param=None):
        """List parameters configured for karbor.

        Lists all parameters configured for karbor unless an optional argument
        is specified.  If the parameter is specified we only print the
        requested parameter.  If the parameter is not found an appropriate
        error is produced by .get*().
        """
        param = param and param.strip()
        if param:
            print('%s = %s' % (param, CONF.get(param)))
        else:
            for key, value in CONF.items():
                print('%s = %s' % (key, value))


class ServiceCommands(object):
    """Methods for managing services."""
    def list(self):
        """Show a list of all karbor services."""

        ctxt = context.get_admin_context()
        services = db.service_get_all(ctxt)
        print_format = "%-16s %-36s %-10s %-5s %-10s"
        print(print_format % (_('Binary'),
                              _('Host'),
                              _('Status'),
                              _('State'),
                              _('Updated At')))
        for svc in services:
            alive = utils.service_is_up(svc)
            art = ":-)" if alive else "XXX"
            status = 'enabled'
            if svc['disabled']:
                status = 'disabled'
            print(print_format % (svc['binary'], svc['host'].partition('.')[0],
                                  status, art,
                                  svc['updated_at']))


CATEGORIES = {
    'config': ConfigCommands,
    'db': DbCommands,
    'service': ServiceCommands,
    'version': VersionCommands,
}


def methods_of(obj):
    """Return non-private methods from an object.

    Get all callable methods of an object that don't start with underscore
    :return: a list of tuples of the form (method_name, method)
    """
    result = []
    for i in dir(obj):
        if callable(getattr(obj, i)) and not i.startswith('_'):
            result.append((i, getattr(obj, i)))
    return result


def add_command_parsers(subparsers):
    for category in CATEGORIES:
        command_object = CATEGORIES[category]()

        parser = subparsers.add_parser(category)
        parser.set_defaults(command_object=command_object)

        category_subparsers = parser.add_subparsers(dest='action')

        for (action, action_fn) in methods_of(command_object):
            parser = category_subparsers.add_parser(action)

            action_kwargs = []
            for args, kwargs in getattr(action_fn, 'args', []):
                parser.add_argument(*args, **kwargs)

            parser.set_defaults(action_fn=action_fn)
            parser.set_defaults(action_kwargs=action_kwargs)


category_opt = cfg.SubCommandOpt('category',
                                 title='Command categories',
                                 handler=add_command_parsers)


def get_arg_string(args):
    arg = None
    if args[0] == '-':
        # (Note)zhiteng: args starts with FLAGS.oparser.prefix_chars
        # is optional args. Notice that cfg module takes care of
        # actual ArgParser so prefix_chars is always '-'.
        if args[1] == '-':
            # This is long optional arg
            arg = args[2:]
        else:
            arg = args[1:]
    else:
        arg = args

    return arg


def fetch_func_args(func):
    fn_args = []
    for args, kwargs in getattr(func, 'args', []):
        arg = get_arg_string(args[0])
        fn_args.append(getattr(CONF.category, arg))

    return fn_args


def main():
    """Parse options and call the appropriate class/method."""
    objects.register_all()
    CONF.register_cli_opt(category_opt)
    script_name = sys.argv[0]
    if len(sys.argv) < 2:
        print(_("\nOpenStack Karbor version: %(version)s\n") %
              {'version': version.version_string()})
        print(script_name + " category action [<args>]")
        print(_("Available categories:"))
        for category in CATEGORIES:
            print(_("\t%s") % category)
        sys.exit(2)

    try:
        CONF(sys.argv[1:], project='karbor',
             version=version.version_string())
        logging.setup(CONF, "karbor")
    except cfg.ConfigDirNotFoundError as details:
        print(_("Invalid directory: %s") % details)
        sys.exit(2)
    except cfg.ConfigFilesNotFoundError:
        cfgfile = CONF.config_file[-1] if CONF.config_file else None
        if cfgfile and not os.access(cfgfile, os.R_OK):
            st = os.stat(cfgfile)
            print(_("Could not read %s, Please try running this"
                    "command again as root/Administrator privilege"
                    "using sudo.") % cfgfile)
            try:
                os.execvp('sudo', ['sudo', '-u', '#%s' % st.st_uid] + sys.argv)
            except Exception:
                print(_('sudo failed, continuing as if nothing happened'))

        print(_('Please re-run karbor-manage as root.'))
        sys.exit(2)

    fn = CONF.category.action_fn
    fn_args = fetch_func_args(fn)
    fn(*fn_args)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from croniter import croniter
from datetime import datetime
from oslo_utils import timeutils

from karbor import exception
from karbor.i18n import _
from karbor.services.operationengine.engine.triggers.timetrigger import \
    timeformats


class Crontab(timeformats.TimeFormat):

    def __init__(self, start_time, pattern):
        self._start_time = start_time
        self._pattern = pattern
        super(Crontab, self).__init__(start_time, pattern)

    @classmethod
    def check_time_format(cls, pattern):
        if not pattern:
            msg = (_("The trigger pattern is None"))
            raise exception.InvalidInput(msg)

        try:
            croniter(pattern)
        except Exception:
            msg = (_("The trigger pattern(%s) is invalid") % pattern)
            raise exception.InvalidInput(msg)

    def compute_next_time(self, current_time):
        time = current_time if current_time >= self._start_time else (
            self._start_time)
        return croniter(self._pattern, time).get_next(datetime)

    def get_min_interval(self):
        try:
            t1 = self.compute_next_time(datetime.now())
            t2 = self.compute_next_time(t1)
            return timeutils.delta_seconds(t1, t2)
        except Exception:
            return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor import exception
from karbor.services.operationengine import operation_manager
from karbor.services.operationengine import operations
from karbor.tests import base


class FakeUserTrustManager(object):
    def add_operation(self, context, operation_id):
        return "123"

    def delete_operation(self, context, operation_id):
        pass

    def resume_operation(self, operation_id, user_id, project_id, trust_id):
        pass


class FakeOperation(operations.base.Operation):
    OPERATION_TYPE = 'fake'

    @classmethod
    def check_operation_definition(cls, operation_definition):
        pass

    @classmethod
    def _execute(cls, operation_definition, param):
        pass

    @classmethod
    def _resume(cls, operation_definition, param, log_ref):
        pass


class OperationManagerTestCase(base.TestCase):
    """Test cases for OperationManager class."""

    def setUp(self):
        super(OperationManagerTestCase, self).setUp()

        self._operation_type = FakeOperation.OPERATION_TYPE
        self._mock_operations = (FakeOperation, )
        with mock.patch(
            'karbor.services.operationengine.operations.all_operations'
        ) as mock_all_ops:
            mock_all_ops.return_value = self._mock_operations
            self._user_trust_manager = FakeUserTrustManager()
            self._op_manager = operation_manager.OperationManager(
                self._user_trust_manager)

    @mock.patch.object(FakeOperation, 'check_operation_definition')
    def test_check_operation_definition(self, mock_check):
        self._op_manager.check_operation_definition(self._operation_type, {})
        mock_check.assert_called_once_with({})

    @mock.patch.object(operations.base.Operation, 'run')
    def test_run_operation(self, mock_run):
        self._op_manager.run_operation(self._operation_type, {})
        mock_run.assert_called_once_with({})

    def test_invalid_operation_type(self):
        self.assertRaisesRegex(exception.InvalidInput,
                               'Invalid operation type:',
                               self._op_manager.check_operation_definition,
                               "123", {})

        self.assertRaisesRegex(exception.InvalidInput,
                               'Invalid operation type:',
                               self._op_manager.run_operation,
                               "123", {})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.api.openstack import wsgi
from karbor.tests import base


class RequestTest(base.TestCase):
    def test_content_type_missing(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.body = b"<body />"
        self.assertIsNone(request.get_content_type())

    def test_content_type_unsupported(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.headers["Content-Type"] = "text/html"
        request.body = b"asdf<br />"
        self.assertIsNone(request.get_content_type())

    def test_content_type_with_charset(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Content-Type"] = "application/json; charset=UTF-8"
        result = request.get_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_accept(self):
        for content_type in ('application/json',):
            request = wsgi.Request.blank('/tests/123')
            request.headers["Accept"] = content_type
            result = request.best_match_content_type()
            self.assertEqual(content_type, result)

    def test_content_type_from_accept_best(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = "application/xml, application/json"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_query_extension(self):
        request = wsgi.Request.blank('/tests/123.json')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

        request = wsgi.Request.blank('/tests/123.invalid')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_accept_default(self):
        request = wsgi.Request.blank('/tests/123.unsupported')
        request.headers["Accept"] = "application/unsupported1"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)


class ActionDispatcherTest(base.TestCase):
    def test_dispatch(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        self.assertEqual('pants', serializer.dispatch({}, action='create'))

    def test_dispatch_action_none(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action=None))

    def test_dispatch_default(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action='update'))


class DictSerializerTest(base.TestCase):
    def test_dispatch_default(self):
        serializer = wsgi.DictSerializer()
        self.assertEqual('', serializer.serialize({}, 'update'))


class JSONDictSerializerTest(base.TestCase):
    def test_json(self):
        input_dict = dict(servers=dict(a=(2, 3)))
        expected_json = '{"servers":{"a":[2,3]}}'
        serializer = wsgi.JSONDictSerializer()
        result = serializer.serialize(input_dict)
        result = result.replace('\n', '').replace(' ', '')
        self.assertEqual(expected_json, result)


class TextDeserializerTest(base.TestCase):
    def test_dispatch_default(self):
        deserializer = wsgi.TextDeserializer()
        self.assertEqual({}, deserializer.deserialize({}, 'update'))


class JSONDeserializerTest(base.TestCase):
    def test_json(self):
        data = """{"a": {
                "a1": "1",
                "a2": "2",
                "bs": ["1", "2", "3", {"c": {"c1": "1"}}],
                "d": {"e": "1"},
                "f": "1"}}"""
        as_dict = {
            'body': {
                'a': {
                    'a1': '1',
                    'a2': '2',
                    'bs': ['1', '2', '3', {'c': {'c1': '1'}}],
                    'd': {'e': '1'},
                    'f': '1',
                },
            },
        }
        deserializer = wsgi.JSONDeserializer()
        self.assertEqual(as_dict, deserializer.deserialize(data))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math
import time

from karbor import exception
from karbor.i18n import _
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import LeasePlugin
from karbor.services.protection import client_factory
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import loopingcall
from oslo_utils import uuidutils
from swiftclient import ClientException


swift_bank_plugin_opts = [
    cfg.StrOpt('bank_swift_object_container',
               default='karbor',
               help='The default swift container to use.'),
]

LOG = logging.getLogger(__name__)

lease_opt = [cfg.IntOpt('lease_expire_window',
                        default=600,
                        help='expired_window for bank lease, in seconds'),
             cfg.IntOpt('lease_renew_window',
                        default=120,
                        help='period for bank lease, in seconds, '
                             'between bank lease client renew the lease'),
             cfg.IntOpt('lease_validity_window',
                        default=100,
                        help='validity_window for bank lease, in seconds'), ]


class SwiftConnectionFailed(exception.KarborException):
    message = _("Connection to swift failed: %(reason)s")


class SwiftBankPlugin(BankPlugin, LeasePlugin):
    """Swift bank plugin"""
    def __init__(self, config, context=None):
        super(SwiftBankPlugin, self).__init__(config)
        self._config.register_opts(swift_bank_plugin_opts,
                                   "swift_bank_plugin")
        self._config.register_opts(lease_opt,
                                   "swift_bank_plugin")
        plugin_cfg = self._config.swift_bank_plugin
        self.bank_object_container = plugin_cfg.bank_swift_object_container
        self.lease_expire_window = plugin_cfg.lease_expire_window
        self.lease_renew_window = plugin_cfg.lease_renew_window
        self.context = context
        # TODO(luobin):
        # init lease_validity_window
        # according to lease_renew_window if not configured
        self.lease_validity_window = plugin_cfg.lease_validity_window

        # TODO(luobin): create a uuid of this bank_plugin
        self.owner_id = uuidutils.generate_uuid()
        self.lease_expire_time = 0
        self.bank_leases_container = "leases"
        self._connection = None

    def _setup_connection(self):
        return client_factory.ClientFactory.create_client('swift',
                                                          self.context,
                                                          self._config)

    @property
    def connection(self):
        if not self._connection:
            _connection = self._setup_connection()
            # create container
            try:
                _connection.put_container(self.bank_object_container)
                _connection.put_container(self.bank_leases_container)
            except SwiftConnectionFailed as err:
                LOG.error("bank plugin create container failed.")
                raise exception.CreateContainerFailed(reason=err)
            self._connection = _connection

            # acquire lease
            try:
                self.acquire_lease()
            except exception.AcquireLeaseFailed:
                LOG.error("bank plugin acquire lease failed.")
                raise

            # start renew lease
            renew_lease_loop = loopingcall.FixedIntervalLoopingCall(
                self.renew_lease)
            renew_lease_loop.start(interval=self.lease_renew_window,
                                   initial_delay=self.lease_renew_window)
        return self._connection

    def get_owner_id(self):
        return self.owner_id

    def update_object(self, key, value):
        serialized = False
        try:
            if not isinstance(value, str):
                value = jsonutils.dumps(value)
                serialized = True
            self._put_object(container=self.bank_object_container,
                             obj=key,
                             contents=value,
                             headers={
                                 'x-object-meta-serialized': str(serialized)
                             })
        except SwiftConnectionFailed as err:
            LOG.error("update object failed, err: %s.", err)
            raise exception.BankUpdateObjectFailed(reason=err, key=key)

    def delete_object(self, key):
        try:
            self._delete_object(container=self.bank_object_container,
                                obj=key)
        except SwiftConnectionFailed as err:
            LOG.error("delete object failed, err: %s.", err)
            raise exception.BankDeleteObjectFailed(reason=err, key=key)

    def get_object(self, key):
        try:
            return self._get_object(container=self.bank_object_container,
                                    obj=key)
        except SwiftConnectionFailed as err:
            LOG.error("get object failed, err: %s.", err)
            raise exception.BankGetObjectFailed(reason=err, key=key)

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        try:
            if sort_dir == "desc":
                body = self._get_container(
                    container=self.bank_object_container,
                    prefix=prefix, end_marker=marker)
                limit_objects = body[-limit:] if limit is not None else body
                return [obj.get("name") for obj in limit_objects]
            else:
                body = self._get_container(
                    container=self.bank_object_container,
                    prefix=prefix, limit=limit, marker=marker)
                return [obj.get("name") for obj in body]
        except SwiftConnectionFailed as err:
            LOG.error("list objects failed, err: %s.", err)
            raise exception.BankListObjectsFailed(reason=err)

    def acquire_lease(self):
        container = self.bank_leases_container
        obj = self.owner_id
        contents = self.owner_id
        headers = {'X-Delete-After': str(self.lease_expire_window)}
        try:
            self._put_object(container=container,
                             obj=obj,
                             contents=contents,
                             headers=headers)
            self.lease_expire_time = math.floor(
                time.time()) + self.lease_expire_window
        except SwiftConnectionFailed as err:
            LOG.error("acquire lease failed, err:%s.", err)
            raise exception.AcquireLeaseFailed(reason=err)

    def renew_lease(self):
        container = self.bank_leases_container
        obj = self.owner_id
        headers = {'X-Delete-After': str(self.lease_expire_window)}
        try:
            self._post_object(container=container,
                              obj=obj,
                              headers=headers)
            self.lease_expire_time = math.floor(
                time.time()) + self.lease_expire_window
        except SwiftConnectionFailed as err:
            LOG.error("acquire lease failed, err:%s.", err)

    def check_lease_validity(self):
        if (self.lease_expire_time - math.floor(time.time()) >=
                self.lease_validity_window):
            return True
        else:
            return False

    def _put_object(self, container, obj, contents, headers=None):
        try:
            self.connection.put_object(container=container,
                                       obj=obj,
                                       contents=contents,
                                       headers=headers)
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)

    def _get_object(self, container, obj):
        try:
            (_resp, body) = self.connection.get_object(container=container,
                                                       obj=obj)
            if _resp.get("x-object-meta-serialized").lower() == "true":
                body = jsonutils.loads(body)
            return body
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)

    def _post_object(self, container, obj, headers):
        try:
            self.connection.post_object(container=container,
                                        obj=obj,
                                        headers=headers)
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)

    def _delete_object(self, container, obj):
        try:
            self.connection.delete_object(container=container,
                                          obj=obj)
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)

    def _put_container(self, container):
        try:
            self.connection.put_container(container=container)
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)

    def _get_container(self, container, prefix=None, limit=None, marker=None,
                       end_marker=None):
        try:
            (_resp, body) = self.connection.get_container(
                container=container,
                prefix=prefix,
                limit=limit,
                marker=marker,
                end_marker=end_marker)
            return body
        except ClientException as err:
            raise SwiftConnectionFailed(reason=err)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import futurist
import six

from karbor import exception
from karbor.i18n import _
from oslo_log import log as logging

from taskflow import engines
from taskflow.patterns import graph_flow
from taskflow.patterns import linear_flow
from taskflow import task


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class WorkFlowEngine(object):
    @abc.abstractmethod
    def build_flow(self, flow_name, flow_type='graph'):
        """build flow

        :param flow_name: the flow name
        :param flow_type:'linear' or 'graph',default:'graph'
        :return:linear flow or graph flow
        """
        return

    @abc.abstractmethod
    def get_engine(self, flow, **kwargs):
        return

    @abc.abstractmethod
    def run_engine(self, flow_engine):
        return

    @abc.abstractmethod
    def output(self, flow_engine, target=None):
        return

    @abc.abstractmethod
    def create_task(self, function, requires=None, provides=None,
                    inject=None, **kwargs):
        """create a task

        :param function:make a task from this callable
        :param requires: A OrderedSet of inputs this task requires to function.
        :param provides:A set, string or list of items that this will be
                         providing (or could provide) to others
        :param inject:An immutable input_name => value dictionary which
                       specifies any initial inputs that should be
                       automatically injected into the task scope before the
                       task execution commences
        """
        return

    @abc.abstractmethod
    def link_task(self, flow, u, v):
        """Link existing node as a runtime dependency of existing node v

        :param u: task or flow to create a link from (must exist already)
        :param v: task or flow to create a link to (must exist already)
        :param flow: graph flow
        """
        return

    @abc.abstractmethod
    def add_tasks(self, flow, *nodes, **kwargs):
        return

    @abc.abstractmethod
    def search_task(self, flow, task_id):
        return


class TaskFlowEngine(WorkFlowEngine):
    def build_flow(self, flow_name, flow_type='graph'):
        if flow_type == 'linear':
            return linear_flow.Flow(flow_name)
        elif flow_type == 'graph':
            return graph_flow.Flow(flow_name)
        else:
            raise ValueError(_("unsupported flow type: %s") % flow_type)

    def get_engine(self, flow, **kwargs):
        if flow is None:
            LOG.error("The flow is None, build it first")
            raise exception.InvalidTaskFlowObject(
                reason=_("The flow is None"))
        executor = kwargs.get('executor', None)
        engine = kwargs.get('engine', None)
        store = kwargs.get('store', None)
        if not executor:
            executor = futurist.GreenThreadPoolExecutor()
        if not engine:
            engine = 'parallel'
        flow_engine = engines.load(flow,
                                   executor=executor,
                                   engine=engine,
                                   store=store)
        return flow_engine

    def karbor_flow_watch(self, state, details):
        LOG.trace("The Flow [%s] OldState[%s] changed to State[%s]: ",
                  details.get('task_name'), details.get('old_state'), state)

    def karbor_atom_watch(self, state, details):
        LOG.trace("The Task [%s] OldState[%s] changed to State[%s]: ",
                  details.get('task_name'), details.get('old_state'), state)

    def run_engine(self, flow_engine):
        if flow_engine is None:
            LOG.error("Flow engine is None,get it first")
            raise exception.InvalidTaskFlowObject(
                reason=_("The flow_engine is None"))

        flow_engine.notifier.register('*', self.karbor_flow_watch)
        flow_engine.atom_notifier.register('*', self.karbor_atom_watch)
        flow_engine.run()

    def output(self, flow_engine, target=None):
        if flow_engine is None:
            LOG.error("Flow engine is None,return nothing")
            raise exception.InvalidTaskFlowObject(
                reason=_("The flow_engine is None"))
        if target:
            return flow_engine.storage.fetch(target)
        return flow_engine.storage.fetch_all()

    def create_task(self, function, requires=None, provides=None,
                    inject=None, **kwargs):
        name = kwargs.get('name', None)
        auto_extract = kwargs.get('auto_extract', True)
        rebind = kwargs.get('rebind', None)
        revert = kwargs.get('revert', None)
        version = kwargs.get('version', None)
        if function:
            return task.FunctorTask(function,
                                    name=name,
                                    provides=provides,
                                    requires=requires,
                                    auto_extract=auto_extract,
                                    rebind=rebind,
                                    revert=revert,
                                    version=version,
                                    inject=inject)

    def link_task(self, flow, u, v):
        if flow is None:
            LOG.error("The flow is None, build it first")
            raise exception.InvalidTaskFlowObject(
                reason=_("The flow is None"))
        if u and v:
            flow.link(u, v)

    def add_tasks(self, flow, *nodes, **kwargs):
        if flow is None:
            LOG.error("The flow is None, get it first")
            raise exception.InvalidTaskFlowObject(
                reason=_("The flow is None"))
        flow.add(*nodes, **kwargs)

    def search_task(self, flow, task_id):
        if not isinstance(flow, graph_flow.Flow):
            LOG.error("this is not a graph flow,flow name:%s", flow.name)
            return
        for node, meta in flow.iter_nodes():
            if not isinstance(node, task.FunctorTask):
                continue
            if task_id == getattr(node, 'name'):
                return node



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta
import eventlet
import functools
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
import six
from stevedore import driver as import_driver

from karbor import exception
from karbor.i18n import _
from karbor.services.operationengine.engine import triggers

time_trigger_opts = [
    cfg.IntOpt('min_interval',
               default=60 * 60,
               help='The minimum interval of two adjacent time points. '
                    'min_interval >= (max_window_time * 2)'),

    cfg.IntOpt('min_window_time',
               default=900,
               help='The minimum window time'),

    cfg.IntOpt('max_window_time',
               default=1800,
               help='The maximum window time'),

    cfg.StrOpt('time_format',
               default='calendar',
               choices=['crontab', 'calendar'],
               help='The type of time format which is used to compute time')
]

CONF = cfg.CONF
CONF.register_opts(time_trigger_opts)
LOG = logging.getLogger(__name__)


class TriggerOperationGreenThread(object):
    def __init__(self, first_run_time, function):
        super(TriggerOperationGreenThread, self).__init__()
        self._is_sleeping = True
        self._pre_run_time = None
        self._running = False
        self._thread = None

        self._function = function

        self._start(first_run_time)

    def kill(self):
        self._running = False
        if self._is_sleeping:
            self._thread.kill()

    @property
    def running(self):
        return self._running

    @property
    def pre_run_time(self):
        return self._pre_run_time

    def _start(self, first_run_time):
        self._running = True

        now = datetime.utcnow()
        initial_delay = 0 if first_run_time <= now else (
            int(timeutils.delta_seconds(now, first_run_time)))

        self._thread = eventlet.spawn_after(
            initial_delay, self._run, first_run_time)
        self._thread.link(self._on_done)

    def _on_done(self, gt, *args, **kwargs):
        self._is_sleeping = True
        self._pre_run_time = None
        self._running = False
        self._thread = None

    def _run(self, expect_run_time):
        while self._running:
            self._is_sleeping = False
            self._pre_run_time = expect_run_time

            expect_run_time = self._function(expect_run_time)
            if expect_run_time is None or not self._running:
                break

            self._is_sleeping = True

            now = datetime.utcnow()
            idle_time = 0 if expect_run_time <= now else int(
                timeutils.delta_seconds(now, expect_run_time))
            eventlet.sleep(idle_time)


class TimeTrigger(triggers.BaseTrigger):
    TRIGGER_TYPE = "time"

    def __init__(self, trigger_id, trigger_property, executor):
        super(TimeTrigger, self).__init__(
            trigger_id, trigger_property, executor)

        self._trigger_property = self.check_trigger_definition(
            trigger_property)

        self._greenthread = None

    def shutdown(self):
        self._kill_greenthread()

    def register_operation(self, operation_id, **kwargs):
        if operation_id in self._operation_ids:
            msg = (_("The operation_id(%s) is exist") % operation_id)
            raise exception.ScheduledOperationExist(msg)

        if self._greenthread and not self._greenthread.running:
            raise exception.TriggerIsInvalid(trigger_id=self._id)

        self._operation_ids.add(operation_id)
        if self._greenthread is None:
            self._start_greenthread()

    def unregister_operation(self, operation_id, **kwargs):
        if operation_id not in self._operation_ids:
            return

        self._operation_ids.remove(operation_id)
        if 0 == len(self._operation_ids):
            self._kill_greenthread()

    def update_trigger_property(self, trigger_property):
        valid_trigger_property = self.check_trigger_definition(
            trigger_property)

        if valid_trigger_property == self._trigger_property:
            return

        timer, first_run_time = self._get_timer_and_first_run_time(
            valid_trigger_property)
        if not first_run_time:
            msg = (_("The new trigger property is invalid, "
                     "Can not find the first run time"))
            raise exception.InvalidInput(msg)

        if self._greenthread is not None:
            pre_run_time = self._greenthread.pre_run_time
            if pre_run_time:
                end_time = pre_run_time + timedelta(
                    seconds=self._trigger_property['window'])
                if first_run_time <= end_time:
                    msg = (_("The new trigger property is invalid, "
                             "First run time%(t1)s must be after %(t2)s") %
                           {'t1': first_run_time, 't2': end_time})
                    raise exception.InvalidInput(msg)

        self._trigger_property = valid_trigger_property

        if len(self._operation_ids) > 0:
            # Restart greenthread to take the change of trigger property
            # effect immediately
            self._kill_greenthread()
            self._create_green_thread(first_run_time, timer)

    def _kill_greenthread(self):
        if self._greenthread:
            self._greenthread.kill()
            self._greenthread = None

    def _start_greenthread(self):
        # Find the first time.
        # We don't known when using this trigger first time.
        timer, first_run_time = self._get_timer_and_first_run_time(
            self._trigger_property)
        if not first_run_time:
            raise exception.TriggerIsInvalid(trigger_id=self._id)

        self._create_green_thread(first_run_time, timer)

    def _create_green_thread(self, first_run_time, timer):
        func = functools.partial(
            self._trigger_operations,
            trigger_property=self._trigger_property.copy(),
            timer=timer)

        self._greenthread = TriggerOperationGreenThread(
            first_run_time, func)

    def _trigger_operations(self, expect_run_time, trigger_property, timer):
        """Trigger operations once

        returns: wait time for next run
        """

        # Just for robustness, actually expect_run_time always <= now
        # but, if the scheduling of eventlet is not accurate, then we
        # can do some adjustments.
        entry_time = datetime.utcnow()
        if entry_time < expect_run_time and (
                int(timeutils.delta_seconds(entry_time, expect_run_time)) > 0):
            return expect_run_time

        # The self._executor.execute_operation may have I/O operation.
        # If it is, this green thread will be switched out during looping
        # operation_ids. In order to avoid changing self._operation_ids
        # during the green thread is switched out, copy self._operation_ids
        # as the iterative object.
        operation_ids = self._operation_ids.copy()
        sent_ops = set()
        window = trigger_property.get("window")
        end_time = expect_run_time + timedelta(seconds=window)

        for operation_id in operation_ids:
            if operation_id not in self._operation_ids:
                # Maybe, when traversing this operation_id, it has been
                # removed by self.unregister_operation
                LOG.warning("Execute operation %s which is not exist, "
                            "ignore it", operation_id)
                continue

            now = datetime.utcnow()
            if now >= end_time:
                LOG.error("Can not trigger operations to run. Because it is "
                          "out of window time. now=%(now)s, "
                          "end time=%(end_time)s, expect run time=%(expect)s,"
                          " wating operations=%(ops)s",
                          {'now': now, 'end_time': end_time,
                           'expect': expect_run_time,
                           'ops': operation_ids - sent_ops})
                break

            try:
                self._executor.execute_operation(
                    operation_id, now, expect_run_time, window)
            except Exception:
                LOG.exception("Submit operation to executor failed, operation"
                              " id=%s", operation_id)

            sent_ops.add(operation_id)

        next_time = self._compute_next_run_time(
            expect_run_time, trigger_property['end_time'], timer)
        now = datetime.utcnow()
        if next_time and next_time <= now:
            LOG.error("Next run time:%(next_time)s <= now:%(now)s. Maybe the "
                      "entry time=%(entry)s is too late, even exceeds the end"
                      " time of window=%(end)s, or it was blocked where "
                      "sending the operation to executor.",
                      {'next_time': next_time, 'now': now,
                       'entry': entry_time, 'end': end_time})
        return next_time

    @classmethod
    def check_trigger_definition(cls, trigger_definition):
        """Check trigger definition

        All the time instances of trigger_definition are in UTC,
        including start_time, end_time
        """
        tf_cls = cls._get_time_format_class()

        pattern = trigger_definition.get("pattern", None)
        tf_cls.check_time_format(pattern)

        start_time = trigger_definition.get("start_time", None)
        if not start_time:
            msg = _("The trigger\'s start time is unknown")
            raise exception.InvalidInput(msg)
        start_time = cls._check_and_get_datetime(start_time, "start_time")

        interval = tf_cls(start_time, pattern).get_min_interval()
        if interval is not None and interval < CONF.min_interval:
            msg = (_("The interval of two adjacent time points "
                     "is less than %d") % CONF.min_interval)
            raise exception.InvalidInput(msg)

        window = trigger_definition.get("window", CONF.min_window_time)
        if not isinstance(window, int):
            try:
                window = int(window)
            except Exception:
                msg = (_("The trigger windows(%s) is not integer") % window)
                raise exception.InvalidInput(msg)

        if window < CONF.min_window_time or window > CONF.max_window_time:
            msg = (_("The trigger windows %(window)d must be between "
                     "%(min_window)d and %(max_window)d") %
                   {"window": window,
                    "min_window": CONF.min_window_time,
                    "max_window": CONF.max_window_time})
            raise exception.InvalidInput(msg)

        end_time = trigger_definition.get("end_time", None)
        end_time = cls._check_and_get_datetime(end_time, "end_time")

        valid_trigger_property = trigger_definition.copy()
        valid_trigger_property['window'] = window
        valid_trigger_property['start_time'] = start_time
        valid_trigger_property['end_time'] = end_time
        return valid_trigger_property

    @classmethod
    def _check_and_get_datetime(cls, time, time_name):
        if not time:
            return None

        if isinstance(time, datetime):
            return time

        if not isinstance(time, six.string_types):
            msg = (_("The trigger %(name)s(type = %(vtype)s) is not an "
                     "instance of string") %
                   {"name": time_name, "vtype": type(time)})
            raise exception.InvalidInput(msg)

        try:
            time = timeutils.parse_strtime(time, fmt='%Y-%m-%d %H:%M:%S')
        except Exception:
            msg = (_("The format of trigger %s is not correct") % time_name)
            raise exception.InvalidInput(msg)

        return time

    @classmethod
    def _compute_next_run_time(cls, start_time, end_time, timer):
        next_time = timer.compute_next_time(start_time)

        if next_time and (not end_time or next_time <= end_time):
            return next_time
        return None

    @classmethod
    def _get_time_format_class(cls):
        return import_driver.DriverManager(
            'karbor.operationengine.engine.timetrigger.time_format',
            CONF.time_format).driver

    @classmethod
    def _get_timer_and_first_run_time(cls, trigger_property):
        tf_cls = cls._get_time_format_class()
        timer = tf_cls(trigger_property['start_time'],
                       trigger_property['pattern'])
        first_run_time = cls._compute_next_run_time(
            datetime.utcnow(), trigger_property['end_time'], timer)

        return timer, first_run_time

    @classmethod
    def check_configuration(cls):
        min_window = CONF.min_window_time
        max_window = CONF.max_window_time
        min_interval = CONF.min_interval

        if not (min_window < max_window and (max_window * 2 <= min_interval)):
            msg = (_('Configurations of time trigger are invalid'))
            raise exception.InvalidInput(msg)



# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the protection manager RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from karbor.objects import base as objects_base
from karbor import rpc


CONF = cfg.CONF


class ProtectionAPI(object):
    """Client side of the protection rpc API.

    API version history:

        1.0 - Initial version.
    """

    RPC_API_VERSION = '1.0'

    def __init__(self):
        super(ProtectionAPI, self).__init__()
        target = messaging.Target(topic=CONF.protection_topic,
                                  version=self.RPC_API_VERSION)
        serializer = objects_base.KarborObjectSerializer()
        self.client = rpc.get_client(target, version_cap=None,
                                     serializer=serializer)

    def restore(self, ctxt, restore=None, restore_auth=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'restore',
            restore=restore,
            restore_auth=restore_auth)

    def protect(self, ctxt, plan=None, checkpoint_properties=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'protect',
            plan=plan,
            checkpoint_properties=checkpoint_properties)

    def delete(self, ctxt, provider_id, checkpoint_id):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'delete',
            provider_id=provider_id,
            checkpoint_id=checkpoint_id)

    def show_checkpoint(self, ctxt, provider_id, checkpoint_id):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'show_checkpoint',
            provider_id=provider_id,
            checkpoint_id=checkpoint_id)

    def list_checkpoints(self, ctxt, provider_id, marker=None,
                         limit=None, sort_keys=None,
                         sort_dirs=None, filters=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'list_checkpoints',
            provider_id=provider_id,
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters)

    def list_protectable_types(self, ctxt):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'list_protectable_types')

    def show_protectable_type(self, ctxt, protectable_type=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'show_protectable_type',
            protectable_type=protectable_type)

    def list_protectable_instances(
            self, ctxt, protectable_type=None,
            marker=None, limit=None, sort_keys=None,
            sort_dirs=None, filters=None, parameters=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'list_protectable_instances',
            protectable_type=protectable_type,
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters,
            parameters=parameters)

    def list_protectable_dependents(self,
                                    ctxt, protectable_id=None,
                                    protectable_type=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'list_protectable_dependents',
            protectable_id=protectable_id,
            protectable_type=protectable_type)

    def show_protectable_instance(self,
                                  ctxt, protectable_type=None,
                                  protectable_id=None,
                                  parameters=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'show_protectable_instance',
            protectable_type=protectable_type,
            protectable_id=protectable_id,
            parameters=parameters)

    def show_provider(self,
                      ctxt, provider_id=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'show_provider',
            provider_id=provider_id)

    def list_providers(self, ctxt, marker=None, limit=None,
                       sort_keys=None,
                       sort_dirs=None, filters=None):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(
            ctxt,
            'list_providers',
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import math
import time

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
from oslo_utils import strutils
import six
import webob

from six.moves import http_client

from karbor import exception
from karbor import i18n
from karbor.i18n import _
from karbor.wsgi import common as wsgi


LOG = logging.getLogger(__name__)

SUPPORTED_CONTENT_TYPES = (
    'application/json',
)

_MEDIA_TYPE_MAP = {
    'application/json': 'json',
}


class Request(wsgi.Request):

    def best_match_content_type(self):
        """Determine the most acceptable content-type.

        Based on:
            1) URI extension (.json)
            2) Content-type header
            3) Accept* headers
        """
        # First lookup http request path
        parts = self.path.rsplit('.', 1)
        if len(parts) > 1:
            _format = parts[1]
            if _format in ['json']:
                return 'application/{0}'.format(_format)

        # Then look up content header
        type_from_header = self.get_content_type()
        if type_from_header:
            return type_from_header
        ctypes = ['application/json']

        # Finally search in Accept-* headers
        bm = self.accept.best_match(ctypes)
        return bm or 'application/json'

    def get_content_type(self):
        allowed_types = ("application/json",)
        if "Content-Type" not in self.headers:
            LOG.debug("Missing Content-Type")
            return None
        _type = self.content_type
        if _type in allowed_types:
            return _type
        return None

    def best_match_language(self):
        """Determines best available locale from the Accept-Language header.

        :returns: the best language match or None if the 'Accept-Language'
                  header was not available in the request.
        """
        if not self.accept_language:
            return None
        all_languages = i18n.get_available_languages()
        return self.accept_language.best_match(all_languages)


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization."""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("cannot understand JSON")
            raise exception.MalformedRequestBody(reason=msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}


class DictSerializer(ActionDispatcher):
    """Default request body serialization."""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return ""


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization."""

    def default(self, data):
        return jsonutils.dumps(data)


def serializers(**serializers):
    """Attaches serializers to a method.

    This decorator associates a dictionary of serializers with a
    method.  Note that the function attributes are directly
    manipulated; the method is not wrapped.
    """

    def decorator(func):
        if not hasattr(func, 'wsgi_serializers'):
            func.wsgi_serializers = {}
        func.wsgi_serializers.update(serializers)
        return func
    return decorator


def deserializers(**deserializers):
    """Attaches deserializers to a method.

    This decorator associates a dictionary of deserializers with a
    method.  Note that the function attributes are directly
    manipulated; the method is not wrapped.
    """

    def decorator(func):
        if not hasattr(func, 'wsgi_deserializers'):
            func.wsgi_deserializers = {}
        func.wsgi_deserializers.update(deserializers)
        return func
    return decorator


def response(code):
    """Attaches response code to a method.

    This decorator associates a response code with a method.  Note
    that the function attributes are directly manipulated; the method
    is not wrapped.
    """

    def decorator(func):
        func.wsgi_code = code
        return func
    return decorator


class ResponseObject(object):
    """Bundles a response object with appropriate serializers.

    Object that app methods may return in order to bind alternate
    serializers with a response object to be serialized.  Its use is
    optional.
    """

    def __init__(self, obj, code=None, **serializers):
        """Binds serializers with an object.

        Takes keyword arguments akin to the @serializer() decorator
        for specifying serializers.  Serializers specified will be
        given preference over default serializers or method-specific
        serializers on return.
        """

        self.obj = obj
        self.serializers = serializers
        self._default_code = http_client.OK
        self._code = code
        self._headers = {}
        self.serializer = None
        self.media_type = None

    def __getitem__(self, key):
        """Retrieves a header with the given name."""

        return self._headers[key.lower()]

    def __setitem__(self, key, value):
        """Sets a header with the given name to the given value."""

        self._headers[key.lower()] = value

    def __delitem__(self, key):
        """Deletes the header with the given name."""

        del self._headers[key.lower()]

    def _bind_method_serializers(self, meth_serializers):
        """Binds method serializers with the response object.

        Binds the method serializers with the response object.
        Serializers specified to the constructor will take precedence
        over serializers specified to this method.

        :param meth_serializers: A dictionary with keys mapping to
                                 response types and values containing
                                 serializer objects.
        """

        # We can't use update because that would be the wrong
        # precedence
        for mtype, serializer in meth_serializers.items():
            self.serializers.setdefault(mtype, serializer)

    def get_serializer(self, content_type, default_serializers=None):
        """Returns the serializer for the wrapped object.

        Returns the serializer for the wrapped object subject to the
        indicated content type.  If no serializer matching the content
        type is attached, an appropriate serializer drawn from the
        default serializers will be used.  If no appropriate
        serializer is available, raises InvalidContentType.
        """

        default_serializers = default_serializers or {}

        try:
            mtype = _MEDIA_TYPE_MAP.get(content_type, content_type)
            if mtype in self.serializers:
                return mtype, self.serializers[mtype]
            else:
                return mtype, default_serializers[mtype]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)

    def preserialize(self, content_type, default_serializers=None):
        """Prepares the serializer that will be used to serialize.

        Determines the serializer that will be used and prepares an
        instance of it for later call.  This allows the serializer to
        be accessed by extensions for, e.g., template extension.
        """

        mtype, serializer = self.get_serializer(content_type,
                                                default_serializers)
        self.media_type = mtype
        self.serializer = serializer()

    def attach(self, **kwargs):
        """Attach slave templates to serializers."""

        if self.media_type in kwargs:
            self.serializer.attach(kwargs[self.media_type])

    def serialize(self, request, content_type, default_serializers=None):
        """Serializes the wrapped object.

        Utility method for serializing the wrapped object.  Returns a
        webob.Response object.
        """

        if self.serializer:
            serializer = self.serializer
        else:
            _mtype, _serializer = self.get_serializer(content_type,
                                                      default_serializers)
            serializer = _serializer()

        response = webob.Response()
        response.status_int = self.code
        for hdr, value in self._headers.items():
            response.headers[hdr] = value
        response.headers['Content-Type'] = content_type
        if self.obj is not None:
            body = serializer.serialize(self.obj)
            if isinstance(body, six.text_type):
                body = body.encode('utf-8')
            response.body = body

        return response

    @property
    def code(self):
        """Retrieve the response status."""

        return self._code or self._default_code

    @property
    def headers(self):
        """Retrieve the headers."""

        return self._headers.copy()


def action_peek_json(body):
    """Determine action to invoke."""

    try:
        decoded = jsonutils.loads(body)
    except ValueError:
        msg = _("cannot understand JSON")
        raise exception.MalformedRequestBody(reason=msg)

    # Make sure there's exactly one key...
    if len(decoded) != 1:
        msg = _("too many body keys")
        raise exception.MalformedRequestBody(reason=msg)

    # Return the action and the decoded body...
    return list(decoded.keys())[0]


class ResourceExceptionHandler(object):
    """Context manager to handle Resource exceptions.

    Used when processing exceptions generated by API implementation
    methods (or their extensions).  Converts most exceptions to Fault
    exceptions, with the appropriate logging.
    """

    def __enter__(self):
        return None

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if not ex_value:
            return True

        if isinstance(ex_value, exception.NotAuthorized):
            raise Fault(webob.exc.HTTPForbidden(explanation=ex_value.msg))
        elif isinstance(ex_value, exception.Invalid):
            raise Fault(exception.ConvertedException(
                code=ex_value.code, explanation=ex_value.msg))
        elif isinstance(ex_value, TypeError):
            exc_info = (ex_type, ex_value, ex_traceback)
            LOG.error('Exception handling resource: %s',
                      ex_value, exc_info=exc_info)
            raise Fault(webob.exc.HTTPBadRequest())
        elif isinstance(ex_value, Fault):
            LOG.info("Fault thrown: %s", ex_value)
            raise ex_value
        elif isinstance(ex_value, webob.exc.HTTPException):
            LOG.info("HTTP exception thrown: %s", ex_value)
            raise Fault(ex_value)

        # We didn't handle the exception
        return False


class Resource(wsgi.Application):
    """WSGI app that handles (de)serialization and controller dispatch.

    WSGI app that reads routing information supplied by RoutesMiddleware
    and calls the requested action method upon its controller.  All
    controller action methods must accept a 'req' argument, which is the
    incoming wsgi.Request. If the operation is a PUT or POST, the controller
    method must also accept a 'body' argument (the deserialized request body).
    They may raise a webob.exc exception or return a dict, which will be
    serialized by requested content type.

    Exceptions derived from webob.exc.HTTPException will be automatically
    wrapped in Fault() to provide API friendly error responses.
    """

    def __init__(self, controller, action_peek=None, **deserializers):
        """Initialize Resource.

        :param controller: object that implement methods created by routes lib
        :param action_peek: dictionary of routines for peeking into an action
                            request body to determine the desired action
        """

        self.controller = controller

        default_deserializers = dict(json=JSONDeserializer)
        default_deserializers.update(deserializers)

        self.default_deserializers = default_deserializers
        self.default_serializers = dict(json=JSONDictSerializer)

        self.action_peek = dict(json=action_peek_json)
        self.action_peek.update(action_peek or {})

        # Copy over the actions dictionary
        self.wsgi_actions = {}
        if controller:
            self.register_actions(controller)

        # Save a mapping of extensions
        self.wsgi_extensions = {}
        self.wsgi_action_extensions = {}

    def register_actions(self, controller):
        """Registers controller actions with this resource."""

        actions = getattr(controller, 'wsgi_actions', {})
        for key, method_name in actions.items():
            self.wsgi_actions[key] = getattr(controller, method_name)

    def register_extensions(self, controller):
        """Registers controller extensions with this resource."""

        extensions = getattr(controller, 'wsgi_extensions', [])
        for method_name, action_name in extensions:
            # Look up the extending method
            extension = getattr(controller, method_name)

            if action_name:
                # Extending an action...
                if action_name not in self.wsgi_action_extensions:
                    self.wsgi_action_extensions[action_name] = []
                self.wsgi_action_extensions[action_name].append(extension)
            else:
                # Extending a regular method
                if method_name not in self.wsgi_extensions:
                    self.wsgi_extensions[method_name] = []
                self.wsgi_extensions[method_name].append(extension)

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""

        # NOTE(Vek): Check for get_action_args() override in the
        # controller
        if hasattr(self.controller, 'get_action_args'):
            return self.controller.get_action_args(request_environment)

        try:
            args = request_environment['wsgiorg.routing_args'][1].copy()
        except (KeyError, IndexError, AttributeError):
            return {}

        try:
            del args['controller']
        except KeyError:
            pass

        try:
            del args['format']
        except KeyError:
            pass

        return args

    def get_body(self, request):

        if len(request.body) == 0:
            LOG.debug("Empty body provided in request")
            return None, ''

        try:
            content_type = request.get_content_type()
        except exception.InvalidContentType:
            LOG.debug("Unrecognized Content-Type provided in request")
            return None, ''

        if not content_type:
            LOG.debug("No Content-Type provided in request")
            return None, ''

        return content_type, request.body

    def deserialize(self, meth, content_type, body):
        meth_deserializers = getattr(meth, 'wsgi_deserializers', {})
        try:
            mtype = _MEDIA_TYPE_MAP.get(content_type, content_type)
            if mtype in meth_deserializers:
                deserializer = meth_deserializers[mtype]
            else:
                deserializer = self.default_deserializers[mtype]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)

        return deserializer().deserialize(body)

    def pre_process_extensions(self, extensions, request, action_args):
        # List of callables for post-processing extensions
        post = []

        for ext in extensions:
            if inspect.isgeneratorfunction(ext):
                response = None

                # If it's a generator function, the part before the
                # yield is the preprocessing stage
                try:
                    with ResourceExceptionHandler():
                        gen = ext(req=request, **action_args)
                        response = next(gen)
                except Fault as ex:
                    response = ex

                # We had a response...
                if response:
                    return response, []

                # No response, queue up generator for post-processing
                post.append(gen)
            else:
                # Regular functions only perform post-processing
                post.append(ext)

        # Run post-processing in the reverse order
        return None, reversed(post)

    def post_process_extensions(self, extensions, resp_obj, request,
                                action_args):
        for ext in extensions:
            response = None
            if inspect.isgenerator(ext):
                # If it's a generator, run the second half of
                # processing
                try:
                    with ResourceExceptionHandler():
                        response = ext.send(resp_obj)
                except StopIteration:
                    # Normal exit of generator
                    continue
                except Fault as ex:
                    response = ex
            else:
                # Regular functions get post-processing...
                try:
                    with ResourceExceptionHandler():
                        response = ext(req=request, resp_obj=resp_obj,
                                       **action_args)
                except Fault as ex:
                    response = ex

            # We had a response...
            if response:
                return response

        return None

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, request):
        """WSGI method that controls (de)serialization and method dispatch."""

        LOG.info("%(method)s %(url)s",
                 {"method": request.method, "url": request.url})

        # Identify the action, its arguments, and the requested
        # content type
        action_args = self.get_action_args(request.environ)
        action = action_args.pop('action', None)
        content_type, body = self.get_body(request)
        accept = request.best_match_content_type()

        # NOTE(Vek): Splitting the function up this way allows for
        #            auditing by external tools that wrap the existing
        #            function.  If we try to audit __call__(), we can
        #            run into troubles due to the @webob.dec.wsgify()
        #            decorator.
        return self._process_stack(request, action, action_args,
                                   content_type, body, accept)

    def _process_stack(self, request, action, action_args,
                       content_type, body, accept):
        """Implement the processing stack."""

        # Get the implementing method
        try:
            meth, extensions = self.get_method(request, action,
                                               content_type, body)
        except (AttributeError, TypeError):
            return Fault(webob.exc.HTTPNotFound())
        except KeyError as ex:
            msg = _("There is no such action: %s") % ex.args[0]
            return Fault(webob.exc.HTTPBadRequest(explanation=msg))
        except exception.MalformedRequestBody:
            msg = _("Malformed request body")
            return Fault(webob.exc.HTTPBadRequest(explanation=msg))

        # Now, deserialize the request body...
        try:
            if content_type:
                contents = self.deserialize(meth, content_type, body)
            else:
                contents = {}
        except exception.InvalidContentType:
            msg = _("Unsupported Content-Type")
            return Fault(webob.exc.HTTPBadRequest(explanation=msg))
        except exception.MalformedRequestBody:
            msg = _("Malformed request body")
            return Fault(webob.exc.HTTPBadRequest(explanation=msg))

        # Update the action args
        action_args.update(contents)

        project_id = action_args.pop("project_id", None)
        context = request.environ.get('karbor.context')
        if (context and project_id and (project_id != context.project_id)):
            msg = _("Malformed request url")
            return Fault(webob.exc.HTTPBadRequest(explanation=msg))

        # Run pre-processing extensions
        response, post = self.pre_process_extensions(extensions,
                                                     request, action_args)

        if not response:
            try:
                with ResourceExceptionHandler():
                    action_result = self.dispatch(meth, request, action_args)
            except Fault as ex:
                response = ex

        if not response:
            # No exceptions; convert action_result into a
            # ResponseObject
            resp_obj = None
            if type(action_result) is dict or action_result is None:
                resp_obj = ResponseObject(action_result)
            elif isinstance(action_result, ResponseObject):
                resp_obj = action_result
            else:
                response = action_result

            # Run post-processing extensions
            if resp_obj:
                _set_request_id_header(request, resp_obj)
                # Do a preserialize to set up the response object
                serializers = getattr(meth, 'wsgi_serializers', {})
                resp_obj._bind_method_serializers(serializers)
                if hasattr(meth, 'wsgi_code'):
                    resp_obj._default_code = meth.wsgi_code
                resp_obj.preserialize(accept, self.default_serializers)

                # Process post-processing extensions
                response = self.post_process_extensions(post, resp_obj,
                                                        request, action_args)

            if resp_obj and not response:
                response = resp_obj.serialize(request, accept,
                                              self.default_serializers)

        try:
            msg_dict = dict(url=request.url, status=response.status_int)
            msg = "%(url)s returned with HTTP %(status)d"
        except AttributeError as e:
            msg_dict = dict(url=request.url, e=e)
            msg = "%(url)s returned a fault: %(e)s"

        LOG.info(msg, msg_dict)

        return response

    def get_method(self, request, action, content_type, body):
        """Look up the action-specific method and its extensions."""

        # Look up the method
        try:
            if not self.controller:
                meth = getattr(self, action)
            else:
                meth = getattr(self.controller, action)
        except AttributeError as e:
            with excutils.save_and_reraise_exception(e) as ctxt:
                if (not self.wsgi_actions or action not in ['action',
                                                            'create',
                                                            'delete',
                                                            'update']):
                    LOG.exception('Get method error.')
                else:
                    ctxt.reraise = False
        else:
            return meth, self.wsgi_extensions.get(action, [])

        if action == 'action':
            # OK, it's an action; figure out which action...
            mtype = _MEDIA_TYPE_MAP.get(content_type)
            action_name = self.action_peek[mtype](body)
            LOG.debug("Action body: %s", body)
        else:
            action_name = action

        # Look up the action method
        return (self.wsgi_actions[action_name],
                self.wsgi_action_extensions.get(action_name, []))

    def dispatch(self, method, request, action_args):
        """Dispatch a call to the action-specific method."""

        return method(req=request, **action_args)


def action(name):
    """Mark a function as an action.

    The given name will be taken as the action key in the body.

    This is also overloaded to allow extensions to provide
    non-extending definitions of create and delete operations.
    """

    def decorator(func):
        func.wsgi_action = name
        return func
    return decorator


def extends(*args, **kwargs):
    """Indicate a function extends an operation.

    Can be used as either::

        @extends
        def index(...):
            pass

    or as::

        @extends(action='resize')
        def _action_resize(...):
            pass
    """

    def decorator(func):
        # Store enough information to find what we're extending
        func.wsgi_extends = (func.__name__, kwargs.get('action'))
        return func

    # If we have positional arguments, call the decorator
    if args:
        return decorator(*args)

    # OK, return the decorator instead
    return decorator


class ControllerMetaclass(type):
    """Controller metaclass.

    This metaclass automates the task of assembling a dictionary
    mapping action keys to method names.
    """

    def __new__(mcs, name, bases, cls_dict):
        """Adds the wsgi_actions dictionary to the class."""

        # Find all actions
        actions = {}
        extensions = []
        # start with wsgi actions from base classes
        for base in bases:
            actions.update(getattr(base, 'wsgi_actions', {}))
        for key, value in cls_dict.items():
            if not callable(value):
                continue
            if getattr(value, 'wsgi_action', None):
                actions[value.wsgi_action] = key
            elif getattr(value, 'wsgi_extends', None):
                extensions.append(value.wsgi_extends)

        # Add the actions and extensions to the class dict
        cls_dict['wsgi_actions'] = actions
        cls_dict['wsgi_extensions'] = extensions

        return super(ControllerMetaclass, mcs).__new__(mcs, name, bases,
                                                       cls_dict)


@six.add_metaclass(ControllerMetaclass)
class Controller(object):
    """Default controller."""

    _view_builder_class = None

    def __init__(self, view_builder=None):
        """Initialize controller with a view builder instance."""
        if view_builder:
            self._view_builder = view_builder
        elif self._view_builder_class:
            self._view_builder = self._view_builder_class()
        else:
            self._view_builder = None

    @staticmethod
    def is_valid_body(body, entity_name):
        if not (body and entity_name in body):
            return False

        def is_dict(d):
            try:
                d.get(None)
                return True
            except AttributeError:
                return False

        if not is_dict(body[entity_name]):
            return False

        return True

    @staticmethod
    def assert_valid_body(body, entity_name):
        # NOTE: After v1 api is deprecated need to merge 'is_valid_body' and
        #       'assert_valid_body' in to one method. Right now it is not
        #       possible to modify 'is_valid_body' to raise exception because
        #       in case of V1 api when 'is_valid_body' return False,
        #       'HTTPUnprocessableEntity' exception is getting raised and in
        #       V2 api 'HTTPBadRequest' exception is getting raised.
        if not Controller.is_valid_body(body, entity_name):
            raise webob.exc.HTTPBadRequest(
                explanation=_("Missing required element '%s' in "
                              "request body.") % entity_name)

    @staticmethod
    def validate_name_and_description(body):
        name = body.get('name')
        if name is not None:
            if isinstance(name, six.string_types):
                body['name'] = name.strip()
            try:
                strutils.check_string_length(body['name'], 'Name',
                                             min_length=0, max_length=255)
            except exception.InvalidInput as error:
                raise webob.exc.HTTPBadRequest(explanation=error.msg)

        description = body.get('description')
        if description is not None:
            try:
                strutils.check_string_length(description, 'Description',
                                             min_length=0, max_length=255)
            except exception.InvalidInput as error:
                raise webob.exc.HTTPBadRequest(explanation=error.msg)

    @staticmethod
    def validate_string_length(value, entity_name, min_length=0,
                               max_length=None, remove_whitespaces=False):
        """Check the length of specified string.

        :param value: the value of the string
        :param entity_name: the name of the string
        :param min_length: the min_length of the string
        :param max_length: the max_length of the string
        :param remove_whitespaces: True if trimming whitespaces is needed
                                   else False
        """
        if isinstance(value, six.string_types) and remove_whitespaces:
            value = value.strip()
        try:
            strutils.check_string_length(value, entity_name,
                                         min_length=min_length,
                                         max_length=max_length)
        except exception.InvalidInput as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)


class Fault(webob.exc.HTTPException):
    """Wrap webob.exc.HTTPException to provide API friendly response."""

    _fault_names = {http_client.BAD_REQUEST: "badRequest",
                    http_client.UNAUTHORIZED: "unauthorized",
                    http_client.FORBIDDEN: "forbidden",
                    http_client.NOT_FOUND: "itemNotFound",
                    http_client.METHOD_NOT_ALLOWED: "badMethod",
                    http_client.CONFLICT: "conflictingRequest",
                    http_client.REQUEST_ENTITY_TOO_LARGE: "overLimit",
                    http_client.UNSUPPORTED_MEDIA_TYPE: "badMediaType",
                    http_client.NOT_IMPLEMENTED: "notImplemented",
                    http_client.SERVICE_UNAVAILABLE: "serviceUnavailable"}

    def __init__(self, exception):
        """Create a Fault for the given webob.exc.exception."""
        self.wrapped_exc = exception
        self.status_int = exception.status_int

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Generate a WSGI response based on the exception passed to ctor."""
        # Replace the body with fault details.
        locale = req.best_match_language()
        code = self.wrapped_exc.status_int
        fault_name = self._fault_names.get(code, "computeFault")
        explanation = self.wrapped_exc.explanation
        fault_data = {
            fault_name: {
                'code': code,
                'message': i18n.translate(explanation, locale)}}
        if code == http_client.REQUEST_ENTITY_TOO_LARGE:
            retry = self.wrapped_exc.headers.get('Retry-After', None)
            if retry:
                fault_data[fault_name]['retryAfter'] = retry

        # 'code' is an attribute on the fault tag itself

        content_type = req.best_match_content_type()
        serializer = {
            'application/json': JSONDictSerializer(),
        }[content_type]

        body = serializer.serialize(fault_data)
        if isinstance(body, six.text_type):
            body = body.encode('utf-8')
        self.wrapped_exc.body = body
        self.wrapped_exc.content_type = content_type
        _set_request_id_header(req, self.wrapped_exc.headers)

        return self.wrapped_exc

    def __str__(self):
        return self.wrapped_exc.__str__()


def _set_request_id_header(req, headers):
    context = req.environ.get('karbor.context')
    if context:
        headers['x-compute-request-id'] = context.request_id


class OverLimitFault(webob.exc.HTTPException):
    """Rate-limited request response."""

    def __init__(self, message, details, retry_time):
        """Initialize new `OverLimitFault` with relevant information."""
        hdrs = OverLimitFault._retry_after(retry_time)
        self.wrapped_exc = webob.exc.HTTPRequestEntityTooLarge(headers=hdrs)
        self.content = {
            "overLimitFault": {
                "code": self.wrapped_exc.status_int,
                "message": message,
                "details": details,
            },
        }

    @staticmethod
    def _retry_after(retry_time):
        delay = int(math.ceil(retry_time - time.time()))
        retry_after = delay if delay > 0 else 0
        headers = {'Retry-After': '%d' % retry_after}
        return headers

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, request):
        """Serializes the wrapped exception conforming to our error format."""
        content_type = request.best_match_content_type()

        def translate(msg):
            locale = request.best_match_language()
            return i18n.translate(msg, locale)

        self.content['overLimitFault']['message'] = translate(
            self.content['overLimitFault']['message'])
        self.content['overLimitFault']['details'] = translate(
            self.content['overLimitFault']['details'])

        serializer = {
            'application/json': JSONDictSerializer(),
        }[content_type]

        content = serializer.serialize(self.content)
        self.wrapped_exc.body = content

        return self.wrapped_exc



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import namedtuple
from oslo_serialization import jsonutils
from oslo_serialization import msgpackutils

from karbor import exception
from karbor import resource
import karbor.services.protection.graph as graph
from karbor.tests import base


class GraphBuilderTest(base.TestCase):
    def test_source_set(self):
        """Test that the source set only contains sources"""

        test_matrix = (
            ({
                "A": ["B"],
                "B": ["C"],
                "C": [],
                }, {"A"}),
            ({
                "A": [],
                "B": ["C"],
                "C": [],
                }, {"A", "B"}),
            ({
                "A": ["C"],
                "B": ["C"],
                "C": [],
                }, {"A", "B"}),
        )

        for g, expected_result in test_matrix:
            result = graph.build_graph(g.keys(), g.__getitem__)
            self.assertEqual(expected_result, {node.value for node in result})

    def test_detect_cyclic_graph(self):
        """Test that cyclic graphs are detected"""

        test_matrix = (
            ({
                "A": ["B"],
                "B": ["C"],
                "C": [],
                }, False),
            ({
                "A": [],
                "B": ["C"],
                "C": [],
                }, False),
            ({
                "A": ["C"],
                "B": ["C"],
                "C": ["A"],
                }, True),
            ({
                "A": ["B"],
                "B": ["C"],
                "C": ["A"],
                }, True),
        )

        for g, expected_result in test_matrix:
            if expected_result:
                self.assertRaises(
                    graph.FoundLoopError,
                    graph.build_graph,
                    g.keys(), g.__getitem__,
                )
            else:
                graph.build_graph(g.keys(), g.__getitem__)

    def test_diamond_graph(self):
        def test_node_children(testnode):
            return testnode.children

        TestNode = namedtuple('TestNode', ['id', 'children'])
#          A
#         / \
#        B   C
#         \ /
#          D
        test_diamond_left = TestNode('D', ())
        test_diamond_right = TestNode('D', ())
        print('id left: ', id(test_diamond_left))
        print('id right:', id(test_diamond_right))
        test_left = TestNode('B', (test_diamond_left, ))
        test_right = TestNode('C', (test_diamond_right, ))
        test_root = TestNode('A', (test_left, test_right, ))
        test_nodes = {test_root, }
        result_graph = graph.build_graph(test_nodes, test_node_children)
        test_root_node = result_graph[0]
        self.assertEqual(2, len(test_root_node.child_nodes))
        test_left_node = test_root_node.child_nodes[0]
        test_right_node = test_root_node.child_nodes[1]

        self.assertEqual(id(test_left_node.child_nodes[0]),
                         id(test_right_node.child_nodes[0]))

    def test_graph_pack_unpack(self):
        test_base = {
            "A1": ["B1", "B2"],
            "B1": ["C1", "C2"],
            "B2": ["C3", "C2"],
            "C1": [],
            "C2": [],
            "C3": [],
        }

        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        packed_graph = graph.pack_graph(test_graph)
        unpacked_graph = graph.unpack_graph(packed_graph)
        self.assertEqual(test_graph, unpacked_graph)

    def test_graph_serialize_deserialize(self):
        Format = namedtuple('Format', ['loads', 'dumps'])
        formats = [
            Format(jsonutils.loads, jsonutils.dumps),
            Format(msgpackutils.loads, msgpackutils.dumps),
        ]
        test_base = {
            "A1": ["B1", "B2"],
            "B1": ["C1", "C2"],
            "B2": ["C3", "C2"],
            "C1": [],
            "C2": [],
            "C3": [],
        }

        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        for fmt in formats:
            serialized = fmt.dumps(graph.pack_graph(test_graph))
            unserialized = graph.unpack_graph(fmt.loads(serialized))
            self.assertEqual(test_graph, unserialized)

    def test_graph_serialize(self):
        resource_a = resource.Resource('server', 0, 'a', {'name': 'a'})
        resource_b = resource.Resource('volume', 1, 'b', {'name': 'b'})
        test_base = {
            resource_a: [resource_b],
            resource_b: []
        }
        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        self.assertIn(
            graph.serialize_resource_graph(test_graph),
            [
                '[{"0x1": ["server", 0, "a", {"name": "a"}], '
                '"0x0": ["volume", 1, "b", {"name": "b"}]}, '
                '[["0x1", ["0x0"]]]]',
                '[{"0x0": ["volume", 1, "b", {"name": "b"}], '
                '"0x1": ["server", 0, "a", {"name": "a"}]}, '
                '[["0x1", ["0x0"]]]]'
            ])

    def test_graph_deserialize_unordered_adjacency(self):
        test_base = {
            "A1": ["B1", "B2"],
            "B1": ["C1", "C2"],
            "B2": ["C3", "C2"],
            "C1": [],
            "C2": [],
            "C3": [],
        }
        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        packed_graph = graph.pack_graph(test_graph)
        reversed_adjacency = tuple(reversed(packed_graph.adjacency))
        packed_graph = graph.PackedGraph(packed_graph.nodes,
                                         reversed_adjacency)
        with self.assertRaisesRegex(exception.InvalidInput, "adjacency list"):
            graph.unpack_graph(packed_graph)

    def test_pack_unpack_graph_with_isolated_node(self):
        test_base = {
            "A1": ["B1", "B2"],
            "B1": ["C1", "C2"],
            "B2": ["C3", "C2"],
            "C1": [],
            "C2": [],
            "C3": [],
            "C4": []
        }

        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        packed_graph = graph.pack_graph(test_graph)
        unpacked_graph = graph.unpack_graph(packed_graph)
        self.assertEqual(len(test_graph), len(unpacked_graph))
        for start_node in test_graph:
            self.assertIn(start_node, unpacked_graph)

    def test_pack_unpack_graph(self):
        test_base = {
            "A1": ["B1", "B2", "B3", "B4"],
            "B1": [],
            "B2": [],
            "B3": ["B1"],
            "B4": ["B2"],
        }

        test_graph = graph.build_graph(test_base.keys(), test_base.__getitem__)
        packed_graph = graph.pack_graph(test_graph)
        unpacked_graph = graph.unpack_graph(packed_graph)
        self.assertEqual(len(test_graph), len(unpacked_graph))
        for start_node in test_graph:
            self.assertIn(start_node, unpacked_graph)


class _TestGraphWalkerListener(graph.GraphWalkerListener):
    def __init__(self, expected_event_stream, test):
        super(_TestGraphWalkerListener, self).__init__()
        # Because the testing framework is badly designed
        # I need to have a reference to the test to raise assertions
        self._test = test
        self._expected_expected_event_stream = list(expected_event_stream)

    def on_node_enter(self, node, already_visited):
        self._test.assertEqual(
            self._expected_expected_event_stream.pop(0),
            ("on_node_enter", node.value, already_visited),
        )

    def on_node_exit(self, node):
        self._test.assertEqual(
            self._expected_expected_event_stream.pop(0),
            ("on_node_exit", node.value),
        )


class GraphWalkerTest(base.TestCase):
    def test_graph_walker(self):
        test_matrix = (
            ({
                'A': ['B'],
                'B': ['C'],
                'C': [],
            }, (
                ("on_node_enter", 'A', False),
                ("on_node_enter", 'B', False),
                ("on_node_enter", 'C', False),
                ("on_node_exit", 'C'),
                ("on_node_exit", 'B'),
                ("on_node_exit", 'A'),
            )),
            ({
                'A': ['C'],
                'B': ['C'],
                'C': [],
            }, (
                ("on_node_enter", 'A', False),
                ("on_node_enter", 'C', False),
                ("on_node_exit", 'C'),
                ("on_node_exit", 'A'),
                ("on_node_enter", 'B', False),
                ("on_node_enter", 'C', True),
                ("on_node_exit", 'C'),
                ("on_node_exit", 'B'),
            )),
            ({
                'A': ['C'],
                'B': ['C'],
                'C': ['D', 'E'],
                'D': [],
                'E': [],
            }, (
                ("on_node_enter", 'A', False),
                ("on_node_enter", 'C', False),
                ("on_node_enter", 'D', False),
                ("on_node_exit", 'D'),
                ("on_node_enter", 'E', False),
                ("on_node_exit", 'E'),
                ("on_node_exit", 'C'),
                ("on_node_exit", 'A'),
                ("on_node_enter", 'B', False),
                ("on_node_enter", 'C', True),
                ("on_node_enter", 'D', True),
                ("on_node_exit", 'D'),
                ("on_node_enter", 'E', True),
                ("on_node_exit", 'E'),
                ("on_node_exit", 'C'),
                ("on_node_exit", 'B'),
            )),
        )

        for g, expected_calls in test_matrix:
            listener = _TestGraphWalkerListener(expected_calls, self)
            walker = graph.GraphWalker()
            walker.register_listener(listener)
            keys = list(g.keys())
            keys.sort()
            walker.walk_graph(graph.build_graph(keys, g.__getitem__))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
from oslo_config import cfg

from webob import exc

from karbor.api.v1 import providers
from karbor import context
from karbor.tests import base
from karbor.tests.unit.api import fakes

CONF = cfg.CONF


class ProvidersApiTest(base.TestCase):
    def setUp(self):
        super(ProvidersApiTest, self).setUp()
        self.controller = providers.ProvidersController()
        self.ctxt = context.RequestContext('demo', 'fakeproject', True)

    @mock.patch(
        'karbor.api.v1.providers.ProvidersController._get_all')
    def test_providers_list_detail(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/providers')
        self.controller.index(req)
        self.assertTrue(moak_get_all.called)

    @mock.patch(
        'karbor.services.protection.api.API.show_provider')
    def test_providers_show(self, moak_show_provider):
        req = fakes.HTTPRequest.blank('/v1/providers')
        self.controller.show(req, '2220f8b1-975d-4621-a872-fa9afb43cb6c')
        self.assertTrue(moak_show_provider.called)

    def test_providers_show_Invalid(self):
        req = fakes.HTTPRequest.blank('/v1/providers')
        self.assertRaises(exc.HTTPBadRequest, self.controller.show,
                          req, "1")

    @mock.patch(
        'karbor.services.protection.api.API.'
        'show_checkpoint')
    def test_checkpoint_show(self, moak_show_checkpoint):
        req = fakes.HTTPRequest.blank('/v1/providers/'
                                      '{provider_id}/checkpoints/')
        moak_show_checkpoint.return_value = {
            "provider_id": "efc6a88b-9096-4bb6-8634-cda182a6e12a",
            "project_id": "446a04d8-6ff5-4e0e-99a4-827a6389e9ff",
            "id": "2220f8b1-975d-4621-a872-fa9afb43cb6c"
        }
        self.controller.checkpoints_show(
            req,
            '2220f8b1-975d-4621-a872-fa9afb43cb6c',
            '2220f8b1-975d-4621-a872-fa9afb43cb6c'
        )
        self.assertTrue(moak_show_checkpoint.called)

    @mock.patch(
        'karbor.services.protection.api.API.'
        'show_checkpoint')
    def test_checkpoint_show_Invalid(self, moak_show_checkpoint):
        req = fakes.HTTPRequest.blank('/v1/providers/'
                                      '{provider_id}/checkpoints/')
        moak_show_checkpoint.return_value = {
            "provider_id": "efc6a88b-9096-4bb6-8634-cda182a6e12a",
            "project_id": "446a04d8-6ff5-4e0e-99a4-827a6389e9ff",
            "id": "2220f8b1-975d-4621-a872-fa9afb43cb6c"
        }
        self.assertRaises(exc.HTTPBadRequest, self.controller.checkpoints_show,
                          req, '2220f8b1-975d-4621-a872-fa9afb43cb6c',
                          '1')

    @mock.patch(
        'karbor.services.protection.api.API.'
        'list_checkpoints')
    def test_checkpoint_index(self, moak_list_checkpoints):
        req = fakes.HTTPRequest.blank('/v1/providers/'
                                      '{provider_id}/checkpoints/')
        moak_list_checkpoints.return_value = [
            {
                "provider_id": "efc6a88b-9096-4bb6-8634-cda182a6e12a",
                "project_id": "446a04d8-6ff5-4e0e-99a4-827a6389e9ff",
                "id": "2220f8b1-975d-4621-a872-fa9afb43cb6c"
            }
        ]
        self.controller.checkpoints_index(
            req,
            '2220f8b1-975d-4621-a872-fa9afb43cb6c')
        self.assertTrue(moak_list_checkpoints.called)

    @mock.patch(
        'karbor.services.protection.api.API.'
        'delete')
    def test_checkpoints_delete(self, moak_delete):
        req = fakes.HTTPRequest.blank('/v1/providers/'
                                      '{provider_id}/checkpoints/')
        self.controller.checkpoints_delete(
            req,
            '2220f8b1-975d-4621-a872-fa9afb43cb6c',
            '2220f8b1-975d-4621-a872-fa9afb43cb6c')
        self.assertTrue(moak_delete.called)

    @mock.patch(
        'karbor.services.protection.api.API.'
        'protect')
    @mock.patch(
        'karbor.objects.plan.Plan.get_by_id')
    def test_checkpoints_create(self, mock_plan_create,
                                mock_protect):
        checkpoint = {
            "plan_id": "2c3a12ee-5ea6-406a-8b64-862711ff85e6"
        }
        body = {"checkpoint": checkpoint}
        req = fakes.HTTPRequest.blank('/v1/providers/'
                                      '{provider_id}/checkpoints/')
        mock_plan_create.return_value = {
            "plan_id": "2c3a12ee-5ea6-406a-8b64-862711ff85e6",
            "provider_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c"
        }
        mock_protect.return_value = {
            "checkpoint_id": "2c3a12ee-5ea6-406a-8b64-862711ff85e6"
        }
        self.controller.checkpoints_create(
            req,
            '2220f8b1-975d-4621-a872-fa9afb43cb6c',
            body)
        self.assertTrue(mock_plan_create.called)
        self.assertTrue(mock_protect.called)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The plans api."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from webob import exc

import karbor
from karbor.api import common
from karbor.api.openstack import wsgi
from karbor.common import constants
from karbor import exception
from karbor.i18n import _

from karbor import objects
from karbor.objects import base as objects_base
import karbor.policy
from karbor.services.operationengine import api as operationengine_api
from karbor.services.protection import api as protection_api
from karbor import utils

import six

query_plan_filters_opt = cfg.ListOpt('query_plan_filters',
                                     default=['name', 'status',
                                              'description'],
                                     help="Plan filter options which "
                                          "non-admin user could use to "
                                          "query plans. Default values "
                                          "are: ['name', 'status', "
                                          "'description']")
CONF = cfg.CONF
CONF.register_opt(query_plan_filters_opt)

LOG = logging.getLogger(__name__)


def check_policy(context, action, target_obj=None):
    target = {
        'project_id': context.project_id,
        'user_id': context.user_id,
    }

    if isinstance(target_obj, objects_base.KarborObject):
        # Turn object into dict so target.update can work
        target.update(
            target_obj.obj_to_primitive() or {})
    else:
        target.update(target_obj or {})

    _action = 'plan:%s' % action
    karbor.policy.enforce(context, _action, target)


class PlanViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "plans"

    def detail(self, request, plan):
        """Detailed view of a single plan."""

        resources = plan.get('resources')
        resources_list = []
        for resource in resources:
            resource_dict = {}
            resource_dict['id'] = resource.pop('id')
            resource_dict['name'] = resource.pop('name')
            resource_dict['type'] = resource.pop('type')
            extra_info = resource.pop('extra_info', None)
            if extra_info:
                resource_dict['extra_info'] = jsonutils.loads(
                    extra_info)
            resources_list.append(resource_dict)
        plan_ref = {
            'plan': {
                'id': plan.get('id'),
                'name': plan.get('name'),
                'description': plan.get('description'),
                'resources': resources_list,
                'provider_id': plan.get('provider_id'),
                'status': plan.get('status'),
                'parameters': plan.get('parameters'),
            }
        }
        return plan_ref

    def detail_list(self, request, plans, plan_count=None):
        """Detailed view of a list of plans."""
        return self._list_view(self.detail, request, plans,
                               plan_count,
                               self._collection_name)

    def _list_view(self, func, request, plans, plan_count,
                   coll_name=_collection_name):
        """Provide a view for a list of plans.

        :param func: Function used to format the plan data
        :param request: API request
        :param plans: List of plans in dictionary format
        :param plan_count: Length of the original list of plans
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: Plan data in dictionary format
        """
        plans_list = [func(request, plan)['plan'] for plan in plans]
        plans_links = self._get_collection_links(request,
                                                 plans,
                                                 coll_name,
                                                 plan_count)
        plans_dict = {}
        plans_dict['plans'] = plans_list
        if plans_links:
            plans_dict['plans_links'] = plans_links

        return plans_dict


class PlansController(wsgi.Controller):
    """The Plans API controller for the OpenStack API."""

    _view_builder_class = PlanViewBuilder

    def __init__(self):
        self.operationengine_api = operationengine_api.API()
        self.protection_api = protection_api.API()
        super(PlansController, self).__init__()

    def show(self, req, id):
        """Return data about the given plan."""
        context = req.environ['karbor.context']

        LOG.info("Show plan with id: %s", id, context=context)

        if not uuidutils.is_uuid_like(id):
            msg = _("Invalid plan id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            plan = self._plan_get(context, id)
        except exception.PlanNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        LOG.info("Show plan request issued successfully.",
                 resource={'id': plan.id})
        return self._view_builder.detail(req, plan)

    def delete(self, req, id):
        """Delete a plan."""
        context = req.environ['karbor.context']

        LOG.info("Delete plan with id: %s", id, context=context)

        try:
            plan = self._plan_get(context, id)
        except exception.PlanNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        check_policy(context, 'delete', plan)
        plan.destroy()
        LOG.info("Delete plan request issued successfully.",
                 resource={'id': plan.id})

    def index(self, req):
        """Returns a list of plans, transformed through view builder."""
        context = req.environ['karbor.context']

        LOG.info("Show plan list", context=context)

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        utils.remove_invalid_filter_options(context,
                                            filters,
                                            self._get_plan_filter_options())

        utils.check_filters(filters)
        plans = self._get_all(context, marker, limit,
                              sort_keys=sort_keys,
                              sort_dirs=sort_dirs,
                              filters=filters,
                              offset=offset)

        retval_plans = self._view_builder.detail_list(req, plans)

        LOG.info("Show plan list request issued successfully.")

        return retval_plans

    def _get_all(self, context, marker=None, limit=None, sort_keys=None,
                 sort_dirs=None, filters=None, offset=None):
        check_policy(context, 'get_all')

        if filters is None:
            filters = {}

        all_tenants = utils.get_bool_param('all_tenants', filters)

        try:
            if limit is not None:
                limit = int(limit)
                if limit < 0:
                    msg = _('limit param must be positive')
                    raise exception.InvalidInput(reason=msg)
        except ValueError:
            msg = _('limit param must be an integer')
            raise exception.InvalidInput(reason=msg)

        if filters:
            LOG.debug("Searching by: %s.", six.text_type(filters))

        if context.is_admin and all_tenants:
            # Need to remove all_tenants to pass the filtering below.
            del filters['all_tenants']
            plans = objects.PlanList.get_all(context, marker, limit,
                                             sort_keys=sort_keys,
                                             sort_dirs=sort_dirs,
                                             filters=filters,
                                             offset=offset)
        else:
            plans = objects.PlanList.get_all_by_project(
                context, context.project_id, marker, limit,
                sort_keys=sort_keys, sort_dirs=sort_dirs, filters=filters,
                offset=offset)

        LOG.info("Get all plans completed successfully.")
        return plans

    def _get_plan_filter_options(self):
        """Return plan search options allowed by non-admin."""
        return CONF.query_plan_filters

    def create(self, req, body):
        """Creates a new plan."""
        if not self.is_valid_body(body, 'plan'):
            raise exc.HTTPUnprocessableEntity()

        LOG.debug('Create plan request body: %s', body)
        context = req.environ['karbor.context']
        check_policy(context, 'create')
        plan = body['plan']
        LOG.debug('Create plan request plan: %s', plan)

        if not plan.get("provider_id"):
            msg = _("provider_id must be provided when creating "
                    "a plan.")
            raise exception.InvalidInput(reason=msg)

        parameters = plan.get("parameters", None)

        if parameters is None:
            msg = _("parameters must be provided when creating "
                    "a plan.")
            raise exception.InvalidInput(reason=msg)

        if not isinstance(parameters, dict):
            msg = _("parameters must be a dict when creating a plan.")
            raise exception.InvalidInput(reason=msg)

        self.validate_name_and_description(plan)
        self.validate_plan_resources(plan)
        self.validate_plan_parameters(context, plan)

        resources = plan.get('resources', None)
        if resources:
            for resource in resources:
                extra_info = resource.get('extra_info', None)
                if extra_info is not None:
                    resource['extra_info'] = jsonutils.dumps(extra_info)

        plan_properties = {
            'name': plan.get('name', None),
            'description': plan.get('description', None),
            'provider_id': plan.get('provider_id', None),
            'project_id': context.project_id,
            'status': constants.PLAN_STATUS_SUSPENDED,
            'resources': resources,
            'parameters': parameters,
        }

        plan = objects.Plan(context=context, **plan_properties)
        plan.create()

        retval = self._view_builder.detail(req, plan)

        return retval

    def update(self, req, id, body):
        """Update a plan."""
        context = req.environ['karbor.context']

        if not body:
            msg = _("Missing request body")
            raise exc.HTTPBadRequest(explanation=msg)

        if 'plan' not in body:
            msg = _("Missing required element '%s' in request body") % 'plan'
            raise exc.HTTPBadRequest(explanation=msg)

        if not uuidutils.is_uuid_like(id):
            msg = _("Invalid plan id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        plan = body['plan']
        update_dict = {}

        valid_update_keys = {
            'name',
            'resources',
            'status',
        }
        for key in valid_update_keys.intersection(plan):
            update_dict[key] = plan[key]

        if update_dict is None:
            msg = _("Missing updated parameters in request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        self.validate_name_and_description(update_dict)
        if update_dict.get("resources"):
            self.validate_plan_resources(update_dict)

        resources = update_dict.get('resources', None)
        if resources:
            for resource in resources:
                extra_info = resource.get('extra_info', None)
                if extra_info is not None:
                    resource['extra_info'] = jsonutils.dumps(extra_info)

        try:
            plan = self._plan_get(context, id)
        except exception.PlanNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        check_policy(context, 'update', plan)
        self._plan_update(context, plan, update_dict)

        plan.update(update_dict)

        retval = self._view_builder.detail(req, plan)
        return retval

    def _plan_get(self, context, plan_id):
        if not uuidutils.is_uuid_like(plan_id):
            msg = _("Invalid plan id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        plan = objects.Plan.get_by_id(context, plan_id)
        try:
            check_policy(context, 'get', plan)
        except exception.PolicyNotAuthorized:
            # raise PlanNotFound instead to make sure karbor behaves
            # as it used to
            raise exception.PlanNotFound(plan_id=plan_id)
        LOG.info("Plan info retrieved successfully.", resource=plan)
        return plan

    def _plan_update(self, context, plan, fields):
        if plan['status'] != constants.PLAN_STATUS_SUSPENDED:
            LOG.info("Unable to update plan, because it is in %s state.",
                     plan['status'])
            msg = _("The plan can be only updated in suspended status.")
            raise exception.InvalidPlan(reason=msg)
        # TODO(chenying) replication scene: need call rpc API when
        # the status of the plan is changed.
        if isinstance(plan, objects_base.KarborObject):
            plan.update(fields)
            plan.save()
            LOG.info("Plan updated successfully.", resource=plan)
        else:
            msg = _("The parameter plan must be a object of "
                    "KarborObject class.")
            raise exception.InvalidInput(reason=msg)

    def validate_plan_resources(self, plan):
        resources_list = plan["resources"]
        if (isinstance(resources_list, list)) and (len(resources_list) > 0):
            for resource in resources_list:
                if (isinstance(resource, dict) and (len(resource) >= 3) and
                        {"id", "type", 'name'}.issubset(resource)):
                    pass
                else:
                    msg = _("Resource in list must be a dict when creating a "
                            "plan.The keys of resource are id,type and name.")
                    raise exception.InvalidInput(reason=msg)
        else:
            msg = _("list resources must be provided when creating "
                    "a plan.")
            raise exception.InvalidInput(reason=msg)

    def validate_plan_parameters(self, context, plan):
        parameters = plan["parameters"]
        if not parameters:
            return
        try:
            provider = self.protection_api.show_provider(
                context, plan["provider_id"])
        except exception:
            msg = _("The provider could not be found.")
            raise exc.HTTPBadRequest(explanation=msg)
        options_schema = provider.get(
            "extended_info_schema", {}).get("options_schema", None)
        if options_schema is None:
            msg = _("The option_schema of plugin must be provided.")
            raise exc.HTTPBadRequest(explanation=msg)
        for resource_key, parameter_value in parameters.items():
            if "#" in resource_key:
                resource_type, resource_id = resource_key.split("#")
                if not uuidutils.is_uuid_like(resource_id):
                    msg = _("The resource_id must be a uuid.")
                    raise exc.HTTPBadRequest(explanation=msg)
            else:
                resource_type = resource_key
            if resource_type not in constants.RESOURCE_TYPES:
                msg = _("The key of plan parameters is invalid.")
                raise exc.HTTPBadRequest(explanation=msg)
            properties = options_schema[resource_type]["properties"]
            if not set(properties.keys()) > set(parameter_value.keys()):
                msg = _("The protect property of plan parameters "
                        "is invalid.")
                raise exc.HTTPBadRequest(explanation=msg)


def create_resource():
    return wsgi.Resource(PlansController())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_versionedobjects import fields

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.objects import base


@base.KarborObjectRegistry.register
class Plan(base.KarborPersistentObject, base.KarborObject,
           base.KarborObjectDictCompat,
           base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    OPTIONAL_FIELDS = ('resources',)

    DEFAULT_EXPECTED_ATTR = ('resources',)

    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(),
        'description': fields.StringField(nullable=True),
        'provider_id': fields.UUIDField(),
        'project_id': fields.UUIDField(),
        'status': fields.StringField(nullable=True),
        'resources': fields.ListOfDictOfNullableStringsField(nullable=False),
        'parameters': base.DictOfDictOfStringsField(),
    }

    # obj_extra_fields is used to hold properties that are not
    # usually part of the model
    obj_extra_fields = ['plan_resources']

    def __init__(self, *args, **kwargs):
        super(Plan, self).__init__(*args, **kwargs)
        self._orig_resources = {}
        self._reset_resources_tracking()

    def obj_reset_changes(self, fields=None):
        super(Plan, self).obj_reset_changes(fields)
        self._reset_resources_tracking(fields=fields)

    def _reset_resources_tracking(self, fields=None):
        if fields is None or 'resources' in fields:
            self._orig_resources = (list(self.resources)
                                    if 'resources' in self else [])

    def obj_what_changed(self):
        changes = super(Plan, self).obj_what_changed()
        if 'resources' in self and self.resources != self._orig_resources:
            changes.add('resources')

        return changes

    @staticmethod
    def _from_db_object(context, plan, db_plan, expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = []
        for name, field in plan.fields.items():
            if name in Plan.OPTIONAL_FIELDS:
                continue
            value = db_plan.get(name)
            if isinstance(field, fields.IntegerField):
                value = value or 0
            if name == "parameters" and value is not None:
                value = jsonutils.loads(value)
            plan[name] = value

        # Get data from db_plan object that was queried by joined query
        # from DB
        if 'resources' in expected_attrs:
            resources = db_plan.get('resources', [])
            resources_list = []
            for resource in resources:
                dict_temp = dict()
                dict_temp['id'] = resource['resource_id']
                dict_temp['type'] = resource['resource_type']
                dict_temp['name'] = resource['resource_name']
                dict_temp['extra_info'] = (
                    resource['resource_extra_info'])
                resources_list.append(dict_temp)
            plan.resources = resources_list

        plan._context = context
        plan.obj_reset_changes()
        return plan

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.karbor_obj_get_changes()

        parameters = updates.pop('parameters', None)
        if parameters is not None:
            updates['parameters'] = jsonutils.dumps(parameters)

        db_plan = db.plan_create(self._context, updates)
        kargs = {}
        if hasattr(Plan, 'DEFAULT_EXPECTED_ATTR'):
            kargs = {'expected_attrs': getattr(Plan, 'DEFAULT_EXPECTED_ATTR')}
        self._from_db_object(self._context, self, db_plan, **kargs)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates:
            if 'parameters' in updates:
                parameters = updates.pop('parameters', None)
                if parameters is not None:
                    updates['parameters'] = jsonutils.dumps(parameters)

            if 'resources' in updates:
                resources = updates.pop('resources', None)
                resources_objlist = db.plan_resources_update(
                    self._context, self.id, resources)
                resources_dictlist = []
                for resource_obj in resources_objlist:
                    resource_dict = {}
                    resource_dict["plan_id"] = resource_obj.get("plan_id")
                    resource_dict["id"] = resource_obj.get("resource_id")
                    resource_dict["type"] = resource_obj.get("resource_type")
                    resource_dict["name"] = resource_obj.get("resource_name")
                    resource_dict["extra_info"] = resource_obj.get(
                        "resource_extra_info")
                    resources_dictlist.append(resource_dict)
                self.resources = resources_dictlist
            db.plan_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        with self.obj_as_admin():
            db.plan_destroy(self._context, self.id)


@base.KarborObjectRegistry.register
class PlanList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('Plan'),
    }

    @base.remotable_classmethod
    def get_all(cls, context, marker, limit, sort_keys=None, sort_dirs=None,
                filters=None, offset=None):
        plans = db.plan_get_all(context, marker, limit,
                                sort_keys=sort_keys, sort_dirs=sort_dirs,
                                filters=filters, offset=offset)
        expected_attrs = ['resources']
        return base.obj_make_list(context, cls(context), objects.Plan,
                                  plans, expected_attrs=expected_attrs)

    @base.remotable_classmethod
    def get_all_by_project(cls, context, project_id, marker, limit,
                           sort_keys=None, sort_dirs=None, filters=None,
                           offset=None):
        plans = db.plan_get_all_by_project(context, project_id, marker,
                                           limit, sort_keys=sort_keys,
                                           sort_dirs=sort_dirs,
                                           filters=filters, offset=offset)
        expected_attrs = ['resources']
        return base.obj_make_list(context, cls(context), objects.Plan,
                                  plans, expected_attrs=expected_attrs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from stevedore import extension as import_driver

from karbor import exception
from karbor.services.operationengine.engine.executors import base as base_exe
from karbor.services.operationengine.engine import triggers
from karbor.services.operationengine.engine.triggers import trigger_manager
from karbor.tests import base


class FakeTrigger(triggers.BaseTrigger):
    def __init__(self, trigger_id, trigger_property, executor):
        super(FakeTrigger, self).__init__(trigger_id, trigger_property,
                                          executor)
        self._ops = set()

    def shutdown(self):
        pass

    def register_operation(self, operation_id, **kwargs):
        self._ops.add(operation_id)

    def unregister_operation(self, operation_id, **kwargs):
        self._ops.discard(operation_id)

    def update_trigger_property(self, trigger_property):
        pass

    @classmethod
    def check_trigger_definition(cls, trigger_definition):
        pass

    @classmethod
    def check_configuration(cls):
        pass

    def has_operations(self):
        return bool(self._ops)


class FakeExecutor(base_exe.BaseExecutor):
    def execute_operation(self, operation_id, triggered_time,
                          expect_start_time, window_time, **kwargs):
        pass

    def resume_operation(self, operation_id, **kwargs):
        pass

    def cancel_operation(self, operation_id):
        pass

    def shutdown(self):
        pass

    @classmethod
    def obj(cls):
        return cls

    @classmethod
    def name(cls):
        return "FakeExecutor"


class TriggerManagerTestCase(base.TestCase):

    def setUp(self):
        super(TriggerManagerTestCase, self).setUp()

        with mock.patch.object(import_driver.ExtensionManager,
                               '_load_plugins') as load_plugin:
            load_plugin.return_value = [FakeExecutor]

            self._executor = FakeExecutor(None)
            self._manager = trigger_manager.TriggerManager(self._executor)
            self._trigger_type = 'fake'
            self._manager._trigger_cls_map[self._trigger_type] = FakeTrigger

    def tearDown(self):
        self._manager.shutdown()
        super(TriggerManagerTestCase, self).tearDown()

    @mock.patch.object(FakeTrigger, 'check_trigger_definition')
    def test_check_trigger_definition(self, func):
        self._manager.check_trigger_definition(self._trigger_type, {})
        func.assert_called_once_with({})

    def test_add_trigger(self):
        trigger_id = 'add'
        self._add_a_trigger(trigger_id)
        self.assertRaisesRegex(exception.InvalidInput,
                               'Trigger id.* is exist',
                               self._manager.add_trigger,
                               trigger_id, self._trigger_type, {})

        self.assertRaisesRegex(exception.InvalidInput,
                               'Invalid trigger type.*',
                               self._manager.add_trigger,
                               1, 'abc', {})

    def test_remove_trigger(self):
        self.assertRaises(exception.TriggerNotFound,
                          self._manager.remove_trigger,
                          1)
        trigger_id = 'remove'
        op_id = 1
        self._add_a_trigger(trigger_id)
        self._manager.register_operation(trigger_id, op_id)

        self.assertRaises(exception.DeleteTriggerNotAllowed,
                          self._manager.remove_trigger,
                          trigger_id)

        self._manager.unregister_operation(trigger_id, op_id)
        self._manager.remove_trigger(trigger_id)
        self.assertRaises(exception.TriggerNotFound,
                          self._manager.remove_trigger,
                          trigger_id)

    @mock.patch.object(FakeTrigger, 'update_trigger_property')
    def test_update_trigger(self, func):
        self.assertRaises(exception.TriggerNotFound,
                          self._manager.update_trigger,
                          1, {})

        trigger_id = 'update'
        self._add_a_trigger(trigger_id)
        self._manager.update_trigger(trigger_id, {})
        func.assert_called_once_with({})

    @mock.patch.object(FakeTrigger, 'register_operation')
    @mock.patch.object(FakeExecutor, 'resume_operation')
    def test_register_operation(self, resume, register):
        self.assertRaises(exception.TriggerNotFound,
                          self._manager.register_operation,
                          1, 1)
        trigger_id = 'register'
        self._add_a_trigger(trigger_id)
        op_id = 1
        self._manager.register_operation(trigger_id, op_id, resume=1)
        register.assert_called_once_with(op_id, resume=1)
        resume.assert_called_once_with(op_id, resume=1)

    @mock.patch.object(FakeTrigger, 'unregister_operation')
    @mock.patch.object(FakeExecutor, 'cancel_operation')
    def test_unregister_operation(self, cancel, unregister):
        self.assertRaises(exception.TriggerNotFound,
                          self._manager.unregister_operation,
                          1, 1)
        trigger_id = 'unregister'
        self._add_a_trigger(trigger_id)
        op_id = 1
        self._manager.unregister_operation(trigger_id, op_id)
        unregister.assert_called_once_with(op_id)
        cancel.assert_called_once_with(op_id)

    def _add_a_trigger(self, trigger_id):
        self._manager.add_trigger(trigger_id, self._trigger_type, {})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

OPTIONS_SCHEMA = {
    "title": "Share Protection Options",
    "type": "object",
    "properties": {
        "snapshot_name": {
            "type": "string",
            "title": "Snapshot Name",
            "description": "The name of the snapshot.",
            "default": None
        },
        "description": {
            "type": "string",
            "title": "Description",
            "description": "The description of the volume."
        },
        "force": {
            "type": "boolean",
            "title": "Force",
            "description": "Whether to backup, even if the volume "
                           "is attached",
            "default": False
        }
    },
    "required": ["snapshot_name", "force"]
}

RESTORE_SCHEMA = {
    "title": "Share Protection Restore",
    "type": "object",
    "properties": {
        "restore_name": {
            "type": "string",
            "title": "Restore Share Name",
            "description": "The name of the restore share",
            "default": None
        },
    },
    "required": ["restore_name"]
}

SAVED_INFO_SCHEMA = {
    "title": "Share Protection Saved Info",
    "type": "object",
    "properties": {},
    "required": []
}
